# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the bounding_box module.
"""

import pytest
from numpy.testing import assert_allclose

from photutils.aperture.bounding_box import BoundingBox
from photutils.aperture.rectangle import RectangularAperture
from photutils.utils._optional_deps import HAS_MATPLOTLIB


def test_bounding_box_init():
    bbox = BoundingBox(1, 10, 2, 20)
    assert bbox.ixmin == 1
    assert bbox.ixmax == 10
    assert bbox.iymin == 2
    assert bbox.iymax == 20


def test_bounding_box_init_minmax():
    match = 'ixmin must be <= ixmax'
    with pytest.raises(ValueError, match=match):
        BoundingBox(100, 1, 1, 100)
    match = 'iymin must be <= iymax'
    with pytest.raises(ValueError, match=match):
        BoundingBox(1, 100, 100, 1)


def test_bounding_box_inputs():
    match = 'ixmin, ixmax, iymin, and iymax must all be integers'
    with pytest.raises(TypeError, match=match):
        BoundingBox([1], [10], [2], [9])
    with pytest.raises(TypeError, match=match):
        BoundingBox([1, 2], 10, 2, 9)
    with pytest.raises(TypeError, match=match):
        BoundingBox(1.0, 10.0, 2.0, 9.0)
    with pytest.raises(TypeError, match=match):
        BoundingBox(1.3, 10, 2, 9)
    with pytest.raises(TypeError, match=match):
        BoundingBox(1, 10.3, 2, 9)
    with pytest.raises(TypeError, match=match):
        BoundingBox(1, 10, 2.3, 9)
    with pytest.raises(TypeError, match=match):
        BoundingBox(1, 10, 2, 9.3)


def test_bounding_box_from_float():
    bbox = BoundingBox.from_float(xmin=1.0, xmax=10.0, ymin=2.0, ymax=20.0)
    assert bbox == BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=21)

    bbox = BoundingBox.from_float(xmin=1.4, xmax=10.4, ymin=1.6, ymax=10.6)
    assert bbox == BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=12)


def test_bounding_box_eq():
    bbox = BoundingBox(1, 10, 2, 20)
    assert bbox == BoundingBox(1, 10, 2, 20)
    assert bbox != BoundingBox(9, 10, 2, 20)
    assert bbox != BoundingBox(1, 99, 2, 20)
    assert bbox != BoundingBox(1, 10, 9, 20)
    assert bbox != BoundingBox(1, 10, 2, 99)

    match = 'Can compare BoundingBox only to another BoundingBox'
    with pytest.raises(TypeError, match=match):
        assert bbox == (1, 10, 2, 20)


def test_bounding_box_repr():
    bbox = BoundingBox(1, 10, 2, 20)
    assert repr(bbox) == 'BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)'


def test_bounding_box_shape():
    bbox = BoundingBox(1, 10, 2, 20)
    assert bbox.shape == (18, 9)


def test_bounding_box_center():
    bbox = BoundingBox(1, 10, 2, 20)
    assert bbox.center == (10.5, 5)


def test_bounding_box_get_overlap_slices():
    bbox = BoundingBox(1, 10, 2, 20)
    slc = ((slice(2, 20, None), slice(1, 10, None)),
           (slice(0, 18, None), slice(0, 9, None)))
    assert bbox.get_overlap_slices((50, 50)) == slc

    bbox = BoundingBox(-10, -1, 2, 20)
    assert bbox.get_overlap_slices((50, 50)) == (None, None)

    bbox = BoundingBox(-10, 10, -10, 20)
    slc = ((slice(0, 20, None), slice(0, 10, None)),
           (slice(10, 30, None), slice(10, 20, None)))
    assert bbox.get_overlap_slices((50, 50)) == slc


def test_bounding_box_get_overlap_slices_invalid_shape():
    """
    Test that get_overlap_slices raises ValueError when shape does not
    have exactly 2 elements.
    """
    bbox = BoundingBox(1, 10, 2, 20)
    match = 'input shape must have 2 elements'
    with pytest.raises(ValueError, match=match):
        bbox.get_overlap_slices((50,))


def test_bounding_box_extent():
    bbox = BoundingBox(1, 10, 2, 20)
    assert_allclose(bbox.extent, (0.5, 9.5, 1.5, 19.5))


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_as_artist():
    bbox = BoundingBox(1, 10, 2, 20)
    patch = bbox.as_artist()

    assert_allclose(patch.get_xy(), (0.5, 1.5))
    assert_allclose(patch.get_width(), 9)
    assert_allclose(patch.get_height(), 18)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_plot():
    from matplotlib.patches import Rectangle
    bbox = BoundingBox(1, 10, 2, 20)
    patch = bbox.plot()
    assert isinstance(patch, Rectangle)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_to_aperture():
    bbox = BoundingBox(1, 10, 2, 20)
    aper = RectangularAperture((5.0, 10.5), w=9.0, h=18.0, theta=0.0)
    bbox_aper = bbox.to_aperture()

    assert_allclose(bbox_aper.positions, aper.positions)
    assert bbox_aper.w == aper.w
    assert bbox_aper.h == aper.h
    assert bbox_aper.theta == aper.theta


def test_bounding_box_union():
    bbox1 = BoundingBox(1, 10, 2, 20)
    bbox2 = BoundingBox(5, 21, 7, 32)
    bbox_union_expected = BoundingBox(1, 21, 2, 32)
    bbox_union1 = bbox1 | bbox2
    bbox_union2 = bbox1.union(bbox2)

    assert bbox_union1 == bbox_union_expected
    assert bbox_union1 == bbox_union2

    match = 'BoundingBox can be joined only with another BoundingBox'
    with pytest.raises(TypeError, match=match):
        bbox1.union((5, 21, 7, 32))


def test_bounding_box_intersect():
    bbox1 = BoundingBox(1, 10, 2, 20)
    bbox2 = BoundingBox(5, 21, 7, 32)
    bbox_intersect_expected = BoundingBox(5, 10, 7, 20)
    bbox_intersect1 = bbox1 & bbox2
    bbox_intersect2 = bbox1.intersection(bbox2)

    assert bbox_intersect1 == bbox_intersect_expected
    assert bbox_intersect1 == bbox_intersect2

    match = 'BoundingBox can be intersected only with another BoundingBox'
    with pytest.raises(TypeError, match=match):
        bbox1.intersection((5, 21, 7, 32))

    assert bbox1.intersection(BoundingBox(30, 40, 50, 60)) is None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for measuring non-parametric morphologies of sources.
"""

import numpy as np

from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['gini']


@deprecated_positional_kwargs(since='3.0', until='4.0')
def gini(data, mask=None):
    r"""
    Calculate the `Gini coefficient
    <https://en.wikipedia.org/wiki/Gini_coefficient>`_ of an array.

    The Gini coefficient of the distribution of absolute flux values
    is calculated using the prescription from `Lotz et al. 2004
    <https://ui.adsabs.harvard.edu/abs/2004AJ....128..163L/abstract>`_
    (Eq. 6) as:

    .. math::

        G = \frac{1}{\overline{|x|} \, n \, (n - 1)}
            \sum^{n}_{i} (2i - n - 1) \left | x_i \right |

    where :math:`\overline{|x|}` is the mean of the absolute value of
    all pixel values :math:`x_i`. If the sum of all pixel values is
    zero, the Gini coefficient is zero.

    The Gini coefficient is a way of measuring the inequality in a given
    set of values. In the context of galaxy morphology, it measures how
    the light of a galaxy image is distributed among its pixels. A Gini
    coefficient value of 0 corresponds to a galaxy image with the light
    evenly distributed over all pixels while a Gini coefficient value of
    1 represents a galaxy image with all its light concentrated in just
    one pixel.

    Usually Gini's measurement needs some sort of preprocessing for
    defining the galaxy region in the image based on the quality of the
    input data. As there is not a general standard for doing this, this
    is left for the user.

    Negative pixel values are used via their absolute value. Invalid
    values (NaN and inf) in the input are automatically excluded from
    the calculation. If only a single finite pixel remains after
    filtering, the Gini coefficient is 0.0.

    Parameters
    ----------
    data : array_like
        The 1D or 2D data array or object that can be converted to an
        array.

    mask : array_like, optional
        A boolean mask with the same shape as ``data`` where `True`
        values indicate masked pixels. Masked pixels are excluded from
        the calculation.

    Returns
    -------
    result : float
        The Gini coefficient of the input array.

    Raises
    ------
    ValueError
        If ``mask`` is provided and does not have the same shape as
        ``data``.
    """
    data = np.asarray(data)
    if mask is not None:
        mask = np.asarray(mask, dtype=bool)
        if mask.shape != data.shape:
            msg = 'mask must have the same shape as data'
            raise ValueError(msg)
        values = np.ravel(data[~mask])
    else:
        values = np.ravel(data)

    # Exclude invalid values (NaN, inf)
    values = np.abs(values[np.isfinite(values)])
    npix = values.size
    if npix == 0:
        return np.nan

    if npix == 1:
        return 0.0

    normalization = np.mean(values) * npix * (npix - 1)
    if normalization == 0.0:
        return 0.0

    kernel = (2.0 * np.arange(1, npix + 1) - npix - 1) * np.sort(values)

    return np.sum(kernel) / normalization



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Documentation build configuration file.

This file is execfile()d with the current directory set to its
containing dir.

Note that not all possible configuration values are present in this
file.

All configuration values have a default. Some values are defined in the
global Astropy configuration which is loaded here before anything else.
See astropy.sphinx.conf for which values are set there.
"""

import os
import sys
import tomllib
from datetime import UTC, datetime
from importlib import metadata
from pathlib import Path

from sphinx.util import logging

logger = logging.getLogger(__name__)

try:
    from sphinx_astropy.conf.v2 import *  # noqa: F403
    from sphinx_astropy.conf.v2 import extensions  # noqa: E402
except ImportError:
    msg = ('The documentation requires the sphinx-astropy package to be '
           'installed. Please install the "docs" requirements.')
    logger.error(msg)
    sys.exit(1)

# Get configuration information from pyproject.toml
with (Path(__file__).parents[1] / 'pyproject.toml').open('rb') as fh:
    project_meta = tomllib.load(fh)['project']

# -- Plot configuration -------------------------------------------------------
plot_rcparams = {
    'axes.labelsize': 'large',
    'figure.figsize': (6, 6),
    'figure.subplot.hspace': 0.5,
    'savefig.bbox': 'tight',
    'savefig.facecolor': 'none',
}
plot_apply_rcparams = True
plot_html_show_source_link = True
plot_formats = ['png', 'hires.png', 'pdf', 'svg']
# Don't use the default - which includes a numpy and matplotlib import
plot_pre_code = ''

# -- General configuration ----------------------------------------------------
# By default, highlight as Python 3.
highlight_language = 'python3'

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '8.2'  # keep in sync with pyproject.toml

# Extend astropy intersphinx_mapping with packages we use here
intersphinx_mapping.update(  # noqa: F405
    {'gwcs': ('https://gwcs.readthedocs.io/en/latest/', None),
     'regions': ('https://astropy-regions.readthedocs.io/en/stable/', None),
     'shapely': ('https://shapely.readthedocs.io/en/stable/', None),
     'skimage': ('https://scikit-image.org/docs/stable/', None),
     })

# Exclude astropy intersphinx_mapping for unused packages
del intersphinx_mapping['h5py']  # noqa: F405

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# .inc.rst mean *include* files, don't have sphinx process them
# exclude_patterns += ["_templates", "_pkgtemplate.rst", "**/*.inc.rst"]

extensions += [
    'sphinx_design',
    'sphinx_reredirects',
]

redirects = {
    'user_guide/epsf': 'epsf_building.html',
    'user_guide/profiles': 'radial_profiles.html',
}

# This is added to the end of RST files - a good place to put
# substitutions to be used globally.
rst_epilog = """
.. _Astropy: https://www.astropy.org/
"""

# -- Project information ------------------------------------------------------
project = project_meta['name']
author = project_meta['authors'][0]['name']
project_copyright = f'2011-{datetime.now(tz=UTC).year}, {author}'
github_project = 'astropy/photutils'

# The version info for the project you're documenting, acts as
# replacement for |version| and |release|, also used in various other
# places throughout the built documents.

# The full version, including alpha/beta/rc tags.
release = metadata.version(project)
# The short X.Y version.
version = '.'.join(release.split('.')[:2])
dev = 'dev' in release

# -- Options for HTML output --------------------------------------------------

html_theme_options = {
    'header_links_before_dropdown': 6,
    'collapse_navigation': True,
    'navigation_depth': 2,
    'show_nav_level': 2,
    'navigation_with_keys': False,
    'use_edit_page_button': False,
    'logo': {
        'image_light': 'photutils_logo_light_plain_path.svg',
        'image_dark': 'photutils_logo_dark_plain_path.svg',
    },
    # alternate way to set the logo
    # 'github_url': 'https://github.com/astropy/photutils',
    'icon_links': [
        {'name': 'GitHub',
         'url': 'https://github.com/astropy/photutils',
         'icon': 'fa-brands fa-github',
         'type': 'fontawesome',
         },
    ],
}

html_title = f'{project} {release}'
html_show_sourcelink = False
html_favicon = os.path.join('_static', 'photutils_logo.ico')
html_static_path = ['_static']
html_css_files = ['custom.css']  # path relative to _static

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get('READTHEDOCS_CANONICAL_URL', '')

# A dictionary of values to pass into the template engine's context for
# all pages.
html_context = {
    'default_mode': 'light',
    'to_be_indexed': ['stable', 'latest'],
    'is_development': dev,
    'github_user': 'astropy',
    'github_repo': 'photutils',
    'github_version': 'main',
    'doc_path': 'docs',
    # Tell Jinja2 templates the build is running on Read the Docs
    'READTHEDOCS': os.environ.get('READTHEDOCS', '') == 'True',
}

# fix size of inheritance diagrams (e.g., PSF diagram was cut off)
inheritance_graph_attrs = {'size': '""'}

# -- Options for LaTeX output -------------------------------------------------
# Grouping the document tree into LaTeX files. List of tuples (source
# start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [('index', project + '.tex', project + ' Documentation',
                    author, 'manual')]
# latex_logo = '_static/photutils_banner.pdf'

# -- Options for manual page output -------------------------------------------
# One entry per manual page. List of tuples (source start file, name,
# description, authors, manual section).
man_pages = [('index', project.lower(), project + ' Documentation',
              [author], 1)]

# -- Resolving issue number to links in changelog -----------------------------
github_issues_url = f'https://github.com/{github_project}/issues/'

# -- Turn on nitpicky mode for sphinx (to warn about references not found) ----
nitpicky = True

# Some warnings are impossible to suppress, and you can list specific
# references that should be ignored in a nitpick-exceptions file which
# should be inside the docs/ directory. The format of the file should be:
#
# <type> <class>
#
# for example:
#
# py:class astropy.io.votable.tree.Element
# py:class astropy.io.votable.tree.SimpleElement
# py:class astropy.io.votable.tree.SimpleElementWithContent
#
# Uncomment the following lines to enable the exceptions:
nitpick_ignore = []
nitpick_filename = 'nitpick-exceptions.txt'
if os.path.isfile(nitpick_filename):
    with open(nitpick_filename) as fh:
        for line in fh:
            if line.strip() == '' or line.startswith('#'):
                continue
            dtype, target = line.split(None, 1)
            target = target.strip()
            nitpick_ignore.append((dtype, target))

# -- Options for linkcheck output ---------------------------------------------
linkcheck_retry = 5
linkcheck_ignore = [
    'http://data.astropy.org',
    r'https://github\.com/astropy/photutils/(?:issues|pull)/\d+',
    # Zenodo/doi: 403 Client Error: Forbidden for url
    r'https://zenodo.org/records/*',
    r'https://doi.org/*',
]
linkcheck_timeout = 180



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.utils._optional_deps import HAS_MATPLOTLIB
from photutils.utils.colormaps import make_random_cmap
from photutils.utils.cutouts import CutoutImage
from photutils.utils.footprints import circular_footprint
from photutils.utils.interpolation import ShepardIDWInterpolator


class TestMakeRandomCmapPositionalKwargs:
    """
    Test make_random_cmap warns for positional optional args.
    """

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_positional_warns(self):
        match = 'make_random_cmap'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            make_random_cmap(100)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_keyword_no_warning(self):
        make_random_cmap(n_colors=100)


class TestCutoutImagePositionalKwargs:
    """
    Test CutoutImage.__init__ warns for positional optional args.
    """

    def setup_method(self):
        self.data = np.arange(100).reshape(10, 10).astype(float)

    def test_positional_warns(self):
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            CutoutImage(self.data, (5, 5), (3, 3), 'trim')

    def test_keyword_no_warning(self):
        CutoutImage(self.data, (5, 5), (3, 3), mode='trim')


class TestCircularFootprintPositionalKwargs:
    """
    Test circular_footprint warns for positional optional args.
    """

    def test_positional_warns(self):
        match = 'circular_footprint'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            circular_footprint(3, float)

    def test_keyword_no_warning(self):
        circular_footprint(3, dtype=float)


class TestShepardIDWInterpolatorPositionalKwargs:
    """
    Test ShepardIDWInterpolator warns for positional optional args.
    """

    def setup_method(self):
        rng = np.random.default_rng(0)
        self.coords = rng.random((100, 2))
        self.values = np.sin(self.coords[:, 0] + self.coords[:, 1])

    def test_init_positional_warns(self):
        weights = np.ones(100)
        match = 'init'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            ShepardIDWInterpolator(self.coords, self.values, weights)

    def test_init_keyword_no_warning(self):
        weights = np.ones(100)
        ShepardIDWInterpolator(self.coords, self.values, weights=weights)

    def test_call_positional_warns(self):
        interp = ShepardIDWInterpolator(self.coords, self.values)
        match = '__call__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            interp([0.5, 0.6], 4)

    def test_call_keyword_no_warning(self):
        interp = ShepardIDWInterpolator(self.coords, self.values)
        interp([0.5, 0.6], n_neighbors=4)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the deprecation of positional optional arguments in the
aperture package.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.aperture.bounding_box import BoundingBox
from photutils.aperture.circle import (CircularAnnulus, CircularAperture,
                                       SkyCircularAnnulus, SkyCircularAperture)
from photutils.aperture.core import SkyAperture
from photutils.aperture.ellipse import (EllipticalAnnulus, EllipticalAperture,
                                        SkyEllipticalAnnulus,
                                        SkyEllipticalAperture)
from photutils.aperture.photometry import aperture_photometry
from photutils.aperture.rectangle import (RectangularAnnulus,
                                          RectangularAperture,
                                          SkyRectangularAnnulus,
                                          SkyRectangularAperture)
from photutils.utils._optional_deps import HAS_MATPLOTLIB

DATA = np.ones((101, 101))
POSITION = (50, 50)
SKY_POSITION = SkyCoord(ra=50.0, dec=50.0, unit='deg')

PIXEL_APERTURE_CL = [CircularAperture,
                     CircularAnnulus,
                     EllipticalAperture,
                     EllipticalAnnulus,
                     RectangularAperture,
                     RectangularAnnulus]

SKY_APERTURE_CL = [SkyCircularAperture,
                   SkyCircularAnnulus,
                   SkyEllipticalAperture,
                   SkyEllipticalAnnulus,
                   SkyRectangularAperture,
                   SkyRectangularAnnulus]

APERTURE_CL = PIXEL_APERTURE_CL + SKY_APERTURE_CL

PIXEL_TEST_APERTURES = list(zip(PIXEL_APERTURE_CL,
                                ({'r': 3.0},
                                 {'r_in': 3.0, 'r_out': 5.0},
                                 {'a': 3.0, 'b': 5.0, 'theta': 1.0},
                                 {'a_in': 3.0, 'a_out': 5.0, 'b_out': 4.0,
                                  'b_in': 12.0 / 5.0, 'theta': 1.0},
                                 {'w': 5, 'h': 8, 'theta': np.pi / 4},
                                 {'w_in': 8, 'w_out': 12, 'h_out': 8,
                                  'h_in': 16.0 / 3.0, 'theta': np.pi / 8}),
                                strict=True))

SKY_TEST_APERTURES = list(zip(SKY_APERTURE_CL,
                              ({'r': 3.0 * u.arcsec},
                               {'r_in': 3.0 * u.arcsec,
                                'r_out': 5.0 * u.arcsec},
                               {'a': 3.0 * u.arcsec, 'b': 5.0 * u.arcsec,
                                'theta': 1.0 * u.deg},
                               {'a_in': 3.0 * u.arcsec,
                                'a_out': 5.0 * u.arcsec,
                                'b_out': 4.0 * u.arcsec,
                                'b_in': 2.4 * u.arcsec,
                                'theta': 1.0 * u.deg},
                               {'w': 5.0 * u.arcsec, 'h': 8.0 * u.arcsec,
                                'theta': 45.0 * u.deg},
                               {'w_in': 8.0 * u.arcsec,
                                'w_out': 12.0 * u.arcsec,
                                'h_out': 8.0 * u.arcsec,
                                'h_in': 16.0 / 3.0 * u.arcsec,
                                'theta': 22.5 * u.deg}),
                              strict=True))

TEST_APERTURES = PIXEL_TEST_APERTURES + SKY_TEST_APERTURES

# Apertures with @deprecated_positional_kwargs on __init__
# (excludes circular apertures which have no optional kwargs)
INIT_WARN_APERTURES = [
    (cls, POSITION, params)
    for cls, params in PIXEL_TEST_APERTURES
    if cls not in (CircularAperture, CircularAnnulus)
] + [
    (cls, SKY_POSITION, params)
    for cls, params in SKY_TEST_APERTURES
    if cls not in (SkyCircularAperture, SkyCircularAnnulus)
]


class TestCircularMaskMixinPositionalKwargs:
    def test_to_mask_no_warning(self):
        aper = CircularAperture(POSITION, r=5)
        aper.to_mask(method='exact', subpixels=5)

    def test_to_mask_positional_warning(self):
        aper = CircularAperture(POSITION, r=5)
        match = 'to_mask'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aper.to_mask('exact')


class TestApertureMethodsPositionalKwargs:
    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_plot_no_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        aper.plot(ax=None)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_plot_positional_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        match = 'plot'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aper.plot(None)

    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_to_mask_no_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        aper.to_mask(method='exact', subpixels=5)

    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_to_mask_positional_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        match = 'to_mask'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aper.to_mask('exact')

    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_do_photometry_no_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        aper.do_photometry(DATA, error=None, mask=None)

    @pytest.mark.parametrize(('aperture_class', 'params'),
                             PIXEL_TEST_APERTURES)
    def test_do_photometry_positional_warning(self, aperture_class, params):
        aper = aperture_class(POSITION, **params)
        error = np.ones_like(DATA)
        match = 'do_photometry'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aper.do_photometry(DATA, error)


class TestApertureInitPositionalKwargs:
    @pytest.mark.parametrize(('aperture_class', 'params'),
                             TEST_APERTURES)
    def test_init_no_warning(self, aperture_class, params):
        position = (SKY_POSITION if issubclass(aperture_class, SkyAperture)
                    else POSITION)
        aperture_class(position, **params)

    @pytest.mark.parametrize(('aperture_class', 'position', 'params'),
                             INIT_WARN_APERTURES)
    def test_init_positional_warning(self, aperture_class, position, params):
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aperture_class(position, *params.values())


class TestBoundingBoxPositionalKwargs:
    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_no_warning(self):
        bbox = BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)
        # Keyword use should not warn
        bbox.plot(ax=None, origin=(0, 0))

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_positional_warning(self):
        bbox = BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)
        match = 'plot'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bbox.plot(None)


class TestApertureMaskPositionalKwargs:
    def setup_method(self):
        aper = CircularAperture(POSITION, r=5)
        self.mask = aper.to_mask(method='exact')

    def test_to_image_no_warning(self):
        self.mask.to_image(shape=DATA.shape, dtype=float)

    def test_to_image_positional_warning(self):
        match = 'to_image'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            self.mask.to_image(DATA.shape, int)

    def test_cutout_no_warning(self):
        self.mask.cutout(DATA, fill_value=0.0, copy=False)

    def test_cutout_positional_warning(self):
        match = 'cutout'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            self.mask.cutout(DATA, 0.0)

    def test_multiply_no_warning(self):
        self.mask.multiply(DATA, fill_value=0.0)

    def test_multiply_positional_warning(self):
        match = 'multiply'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            self.mask.multiply(DATA, 0.0)

    def test_get_values_no_warning(self):
        self.mask.get_values(DATA, mask=None)

    def test_get_values_positional_warning(self):
        match = 'get_values'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            self.mask.get_values(DATA, None)


class TestAperturePhotometryPositionalKwargs:
    def test_no_warning(self):
        aper = CircularAperture(POSITION, r=5)
        aperture_photometry(DATA, aper, error=None, mask=None)

    def test_positional_warning(self):
        aper = CircularAperture(POSITION, r=5)
        error = np.ones_like(DATA)
        match = 'aperture_photometry'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            aperture_photometry(DATA, aper, error)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the flags module.
"""

import numpy as np
import pytest

from photutils.psf import IterativePSFPhotometry, PSFPhotometry
from photutils.psf.flags import (PSF_FLAGS, _PSFFlagDefinition, _PSFFlags,
                                 _update_decode_docstring, decode_psf_flags)


def test_decode_psf_flags():
    """
    Test the decode_psf_flags standalone function.
    """
    # Test single flag value with no flags set
    decoded = decode_psf_flags(0)
    assert decoded == []
    assert isinstance(decoded, list)

    # Test single flag value with one bit set
    decoded = decode_psf_flags(1)
    assert decoded == ['n_pixels_fit_partial']

    decoded = decode_psf_flags(2)
    assert decoded == ['outside_bounds']

    decoded = decode_psf_flags(4)
    assert decoded == ['negative_flux']

    decoded = decode_psf_flags(8)
    assert decoded == ['no_convergence']

    decoded = decode_psf_flags(16)
    assert decoded == ['no_covariance']

    decoded = decode_psf_flags(32)
    assert decoded == ['near_bound']

    decoded = decode_psf_flags(64)
    assert decoded == ['no_overlap']

    decoded = decode_psf_flags(128)
    assert decoded == ['fully_masked']

    decoded = decode_psf_flags(256)
    assert decoded == ['too_few_pixels']

    decoded = decode_psf_flags(512)
    assert decoded == ['non_finite_position']

    decoded = decode_psf_flags(1024)
    assert decoded == ['non_finite_flux']

    decoded = decode_psf_flags(2048)
    assert decoded == ['non_finite_localbkg']

    # Test combination of flags
    decoded = decode_psf_flags(5)  # bits 1 and 4
    assert set(decoded) == {'n_pixels_fit_partial', 'negative_flux'}
    assert len(decoded) == 2

    decoded = decode_psf_flags(136)  # bits 8 and 128
    assert set(decoded) == {'no_convergence', 'fully_masked'}
    assert len(decoded) == 2

    # Test with all flags set
    all_flags = (1 + 2 + 4 + 8 + 16 + 32 + 64 + 128 + 256 + 512 + 1024
                 + 2048)  # 4095
    decoded = decode_psf_flags(all_flags)
    expected_all = ['n_pixels_fit_partial', 'outside_bounds', 'negative_flux',
                    'no_convergence', 'no_covariance', 'near_bound',
                    'no_overlap', 'fully_masked', 'too_few_pixels',
                    'non_finite_position', 'non_finite_flux',
                    'non_finite_localbkg']
    assert set(decoded) == set(expected_all)
    assert len(decoded) == 12

    # Test with array input
    flags_array = [0, 1, 2, 5]
    decoded_list = decode_psf_flags(flags_array)
    assert len(decoded_list) == 4
    assert isinstance(decoded_list, list)

    # Check individual results
    assert decoded_list[0] == []
    assert decoded_list[1] == ['n_pixels_fit_partial']
    assert decoded_list[2] == ['outside_bounds']
    assert set(decoded_list[3]) == {'n_pixels_fit_partial', 'negative_flux'}

    # Test with numpy array
    flags_np = np.array([8, 16, 32])
    decoded_list = decode_psf_flags(flags_np)
    assert len(decoded_list) == 3
    assert decoded_list[0] == ['no_convergence']
    assert decoded_list[1] == ['no_covariance']
    assert decoded_list[2] == ['near_bound']

    # Test with 0-d numpy array (scalar array)
    flag_scalar = np.array(64)
    decoded = decode_psf_flags(flag_scalar)
    assert isinstance(decoded, list)
    assert decoded == ['no_overlap']

    # Test membership operations (common usage pattern)
    issues = decode_psf_flags(136)
    assert 'no_convergence' in issues
    assert 'fully_masked' in issues
    assert 'negative_flux' not in issues

    # Test error conditions
    match = 'Flag value must be an integer'
    with pytest.raises(TypeError, match=match):
        decode_psf_flags(3.14)
    with pytest.raises(TypeError, match=match):
        decode_psf_flags('invalid')
    with pytest.raises(TypeError, match=match):
        decode_psf_flags([1, 2.5, 3])


def test_decode_psf_flags_practical_usage():
    """
    Test practical usage patterns for decode_psf_flags.
    """
    # Simulate some typical flag values
    typical_flags = [0, 1, 8, 64, 136, 256]

    # Test batch processing
    all_issues = decode_psf_flags(typical_flags)
    assert len(all_issues) == len(typical_flags)

    # Test filtering for specific conditions
    convergence_issues = [i for i, issues in enumerate(all_issues)
                          if 'no_convergence' in issues]
    expected_conv_indices = [2, 4]  # flags 8 and 136 have convergence issues
    assert convergence_issues == expected_conv_indices

    # Test counting issues
    issue_counts = {}
    for issues in all_issues:
        for issue in issues:
            issue_counts[issue] = issue_counts.get(issue, 0) + 1

    # Verify expected counts
    assert issue_counts.get('no_convergence', 0) == 2  # flags 8 and 136
    assert issue_counts.get('fully_masked', 0) == 1  # flag 136
    assert issue_counts.get('n_pixels_fit_partial', 0) == 1  # flag 1

    # Test boolean context (empty list is falsy)
    clean_sources = [i for i, issues in enumerate(all_issues) if not issues]
    assert 0 in clean_sources  # flag 0 should have no issues

    # Test string formatting for reporting
    for i, issues in enumerate(all_issues):
        if issues:
            report = f"Source {i}: {', '.join(issues)}"
            assert isinstance(report, str)
            assert str(i) in report


def test_decode_psf_flags_edge_cases():
    """
    Test edge cases for decode_psf_flags.
    """
    # Test with very large flag value (all bits set + extra)
    large_flag = 2**16 - 1  # Much larger than our defined flags
    decoded = decode_psf_flags(large_flag)
    expected_all = ['n_pixels_fit_partial', 'outside_bounds', 'negative_flux',
                    'no_convergence', 'no_covariance', 'near_bound',
                    'no_overlap', 'fully_masked', 'too_few_pixels',
                    'non_finite_position', 'non_finite_flux',
                    'non_finite_localbkg']
    assert set(decoded) == set(expected_all)

    match = 'Flag value must be a non-negative integer'
    with pytest.raises(ValueError, match=match):
        decode_psf_flags(-2)

    # Test with empty array
    empty_array = np.array([], dtype=int)
    decoded = decode_psf_flags(empty_array)
    assert decoded == []

    # Test with 2D array (should flatten)
    flag_2d = np.array([[0, 1], [8, 136]])
    decoded = decode_psf_flags(flag_2d)
    assert len(decoded) == 4  # Flattened to 4 elements
    assert decoded[0] == []
    assert decoded[1] == ['n_pixels_fit_partial']
    assert decoded[2] == ['no_convergence']
    assert set(decoded[3]) == {'no_convergence', 'fully_masked'}


def test_psf_flags_singleton():
    """
    Test _PSFFlags singleton behavior.
    """
    # Test that PSF_FLAGS is accessible and is a _PSFFlags instance
    assert isinstance(PSF_FLAGS, _PSFFlags)

    # Test that multiple references point to the same object
    flags1 = PSF_FLAGS
    flags2 = PSF_FLAGS
    assert flags1 is flags2

    # Test that creating a new instance works independently
    new_flags = _PSFFlags()
    assert isinstance(new_flags, _PSFFlags)
    assert new_flags is not PSF_FLAGS  # Different instances


def test_psf_flags_constants():
    """
    Test _PSFFlags constant access.
    """
    # Test all flag constants exist and have correct values
    expected_constants = {
        'N_PIXELS_FIT_PARTIAL': 1,
        'OUTSIDE_BOUNDS': 2,
        'NEGATIVE_FLUX': 4,
        'NO_CONVERGENCE': 8,
        'NO_COVARIANCE': 16,
        'NEAR_BOUND': 32,
        'NO_OVERLAP': 64,
        'FULLY_MASKED': 128,
        'TOO_FEW_PIXELS': 256,
        'NON_FINITE_POSITION': 512,
        'NON_FINITE_FLUX': 1024,
        'NON_FINITE_LOCALBKG': 2048,
    }

    for const_name, expected_value in expected_constants.items():
        assert hasattr(PSF_FLAGS, const_name)
        actual_value = getattr(PSF_FLAGS, const_name)
        assert actual_value == expected_value
        assert isinstance(actual_value, int)


def test_psf_flags_properties():
    """
    Test _PSFFlags property access methods.
    """
    # Test bit_values property
    bit_values = PSF_FLAGS.bit_values
    expected_bits = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048]
    assert set(bit_values) == set(expected_bits)
    assert len(bit_values) == 12

    # Test names property
    names = PSF_FLAGS.names
    expected_names = [
        'n_pixels_fit_partial', 'outside_bounds', 'negative_flux',
        'no_convergence', 'no_covariance', 'near_bound',
        'no_overlap', 'fully_masked', 'too_few_pixels',
        'non_finite_position', 'non_finite_flux', 'non_finite_localbkg',
    ]
    assert set(names) == set(expected_names)
    assert len(names) == 12

    # Test flag_dict property
    flag_dict = PSF_FLAGS.flag_dict
    assert isinstance(flag_dict, dict)
    assert len(flag_dict) == 12
    for bit_val, name in flag_dict.items():
        assert bit_val in expected_bits
        assert name in expected_names

    # Test all_flags property
    all_flags = PSF_FLAGS.all_flags
    assert isinstance(all_flags, list)
    assert len(all_flags) == 12
    for flag_def in all_flags:
        assert isinstance(flag_def, _PSFFlagDefinition)


def test_psf_flags_get_methods():
    """
    Test _PSFFlags getter methods.
    """
    # Test get_name
    assert PSF_FLAGS.get_name(1) == 'n_pixels_fit_partial'
    assert PSF_FLAGS.get_name(8) == 'no_convergence'
    assert PSF_FLAGS.get_name(256) == 'too_few_pixels'
    assert PSF_FLAGS.get_name(512) == 'non_finite_position'
    assert PSF_FLAGS.get_name(1024) == 'non_finite_flux'
    assert PSF_FLAGS.get_name(2048) == 'non_finite_localbkg'

    # Test get_bit_value
    assert PSF_FLAGS.get_bit_value('n_pixels_fit_partial') == 1
    assert PSF_FLAGS.get_bit_value('no_convergence') == 8
    assert PSF_FLAGS.get_bit_value('too_few_pixels') == 256

    # Test get_description
    desc1 = PSF_FLAGS.get_description(1)
    assert 'n_pixels_fit smaller than full fit_shape region' in desc1

    desc8 = PSF_FLAGS.get_description(8)
    assert 'possible non-convergence' in desc8

    # Test get_detailed_description
    detailed1 = PSF_FLAGS.get_detailed_description(1)
    assert 'number of fitted pixels' in detailed1
    assert 'partial PSF fitting' in detailed1

    detailed8 = PSF_FLAGS.get_detailed_description(8)
    assert 'algorithm may not have converged' in detailed8


def test_psf_flags_get_definition():
    """
    Test _PSFFlags get_definition method.
    """
    # Test get_definition by bit value
    def_by_bit = PSF_FLAGS.get_definition(1)
    assert isinstance(def_by_bit, _PSFFlagDefinition)
    assert def_by_bit.bit_value == 1
    assert def_by_bit.name == 'n_pixels_fit_partial'

    # Test get_definition by name
    def_by_name = PSF_FLAGS.get_definition('n_pixels_fit_partial')
    assert isinstance(def_by_name, _PSFFlagDefinition)
    assert def_by_name.bit_value == 1
    assert def_by_name.name == 'n_pixels_fit_partial'

    # Test that both methods return the same object
    assert def_by_bit is def_by_name

    # Test error cases
    match = 'No flag with bit value 999'
    with pytest.raises(KeyError, match=match):
        PSF_FLAGS.get_definition(999)

    match = "No flag with name 'invalid'"
    with pytest.raises(KeyError, match=match):
        PSF_FLAGS.get_definition('invalid')

    match = 'identifier must be int'
    with pytest.raises(TypeError, match=match):
        PSF_FLAGS.get_definition(3.14)


def test_psf_flag_definition():
    """
    Test _PSFFlagDefinition dataclass.
    """
    # Create a flag definition
    flag_def = _PSFFlagDefinition(
        bit_value=1,
        name='test_flag',
        description='test description',
        detailed_description='detailed test description',
    )

    # Test attributes
    assert flag_def.bit_value == 1
    assert flag_def.name == 'test_flag'
    assert flag_def.description == 'test description'
    assert flag_def.detailed_description == 'detailed test description'

    # Test immutability (frozen dataclass)
    with pytest.raises(AttributeError):
        flag_def.bit_value = 2

    # Test equality
    flag_def2 = _PSFFlagDefinition(
        bit_value=1,
        name='test_flag',
        description='test description',
        detailed_description='detailed test description',
    )
    assert flag_def == flag_def2

    # Test inequality
    flag_def3 = _PSFFlagDefinition(
        bit_value=2,
        name='test_flag',
        description='test description',
        detailed_description='detailed test description',
    )
    assert flag_def != flag_def3


def test_psf_flags_integration_with_decode():
    """
    Test integration between _PSFFlags and decode_psf_flags.
    """
    # Test that decode_psf_flags uses PSF_FLAGS internally
    test_flags = [PSF_FLAGS.N_PIXELS_FIT_PARTIAL, PSF_FLAGS.NO_CONVERGENCE,
                  PSF_FLAGS.FULLY_MASKED]

    decoded = decode_psf_flags(test_flags)
    assert len(decoded) == 3
    assert decoded[0] == ['n_pixels_fit_partial']
    assert decoded[1] == ['no_convergence']
    assert decoded[2] == ['fully_masked']

    # Test combined flags
    combined = PSF_FLAGS.NO_CONVERGENCE | PSF_FLAGS.FULLY_MASKED
    decoded_combined = decode_psf_flags(combined)
    assert set(decoded_combined) == {'no_convergence', 'fully_masked'}

    # Test all constants work with decode
    for const_name in ['N_PIXELS_FIT_PARTIAL', 'OUTSIDE_BOUNDS',
                       'NEGATIVE_FLUX',
                       'NO_CONVERGENCE', 'NO_COVARIANCE', 'NEAR_BOUND',
                       'NO_OVERLAP', 'FULLY_MASKED', 'TOO_FEW_PIXELS']:
        const_value = getattr(PSF_FLAGS, const_name)
        decoded_const = decode_psf_flags(const_value)
        assert len(decoded_const) == 1

        # The decoded name should match the constant name (lowercase)
        expected_name = const_name.lower()
        assert decoded_const[0] == expected_name


def test_psf_flags_completeness():
    """
    Test that _PSFFlags covers all expected flag scenarios.
    """
    # Test that we have the expected number of flags
    assert len(PSF_FLAGS.all_flags) == 12

    # Test that bit values are powers of 2
    for bit_val in PSF_FLAGS.bit_values:
        assert bit_val > 0
        assert (bit_val & (bit_val - 1)) == 0  # Power of 2 check

    # Test that bit values are unique
    bit_values = PSF_FLAGS.bit_values
    assert len(bit_values) == len(set(bit_values))

    # Test that names are unique
    names = PSF_FLAGS.names
    assert len(names) == len(set(names))

    # Test that all names are valid Python identifiers (for compatibility)
    for name in names:
        assert name.isidentifier()
        assert '_' in name or name.islower()  # Snake_case convention

    # Test that all flags can be combined without conflicts
    all_combined = 0
    for bit_val in PSF_FLAGS.bit_values:
        all_combined |= bit_val

    decoded_all = decode_psf_flags(all_combined)
    assert len(decoded_all) == 12
    assert set(decoded_all) == set(PSF_FLAGS.names)


def test_psf_classes_docstrings():
    """
    Test that the PSF classes have dynamic flag documentation.
    """
    classes_to_test = [PSFPhotometry, IterativePSFPhotometry]

    for cls in classes_to_test:
        docstring = cls.__call__.__doc__

        # Should have flags section
        assert '* ``flags`` : bitwise flag values' in docstring

        # Should have all dynamic flag descriptions
        dynamic_flags = [
            'n_pixels_fit smaller than full fit_shape region',
            'fitted position outside input image bounds',
            'non-positive flux',
            'possible non-convergence',
            'missing parameter covariance',
            'fitted parameter near a bound',
            'no overlap with data',
            'fully masked source',
            'too few pixels for fitting',
        ]

        for flag_desc in dynamic_flags:
            msg = f'Missing flag description in {cls.__name__}: {flag_desc}'
            assert flag_desc in docstring, msg


def test_decode_psf_flags_docstring():
    """
    Test that the decode_psf_flags function has dynamic flag
    documentation.
    """
    docstring = decode_psf_flags.__doc__

    # Should not have placeholder
    assert '<flag descriptions>' not in docstring

    # Should have all expected flag names in the expected format
    expected_flags = [
        "``'n_pixels_fit_partial'`` : bit 1",
        "``'outside_bounds'`` : bit 2",
        "``'negative_flux'`` : bit 4",
        "``'no_convergence'`` : bit 8",
        "``'no_covariance'`` : bit 16",
        "``'near_bound'`` : bit 32",
        "``'no_overlap'`` : bit 64",
        "``'fully_masked'`` : bit 128",
        "``'too_few_pixels'`` : bit 256",
    ]

    for flag_desc in expected_flags:
        msg = f'Missing flag in docstring: {flag_desc}'
        assert flag_desc in docstring, msg

    # Should have flag descriptions
    expected_descriptions = [
        'n_pixels_fit smaller than full fit_shape region',
        'fitted position outside input image bounds',
        'non-positive flux',
        'possible non-convergence',
        'missing parameter covariance',
        'fitted parameter near a bound',
        'no overlap with data',
        'fully masked source',
        'too few pixels for fitting',
    ]

    for desc in expected_descriptions:
        assert desc in docstring, f'Missing description: {desc}'


def test_update_decode_docstring_noop():
    """
    Test that the update_decode_docstring decorator is a no-op if no
    docstring exists.
    """
    @_update_decode_docstring
    def test_func(data):
        pass

    docstring = test_func.__doc__
    assert docstring is None


def test_decode_psf_flags_return_bit_values():
    """
    Test the decode_psf_flags function with return_bit_values=True.
    """
    # Test single flag value with no flags set
    decoded = decode_psf_flags(0, return_bit_values=True)
    assert decoded == []
    assert isinstance(decoded, list)

    # Test single flag value with one bit set
    decoded = decode_psf_flags(1, return_bit_values=True)
    assert decoded == [1]

    decoded = decode_psf_flags(2, return_bit_values=True)
    assert decoded == [2]

    # Test combination of flags
    decoded = decode_psf_flags(5, return_bit_values=True)  # bits 1 and 4
    assert set(decoded) == {1, 4}
    assert len(decoded) == 2

    decoded = decode_psf_flags(136, return_bit_values=True)  # bits 8 and 128
    assert set(decoded) == {8, 128}
    assert len(decoded) == 2

    # Test with all flags set
    all_flags = (1 + 2 + 4 + 8 + 16 + 32 + 64 + 128 + 256 + 512 + 1024
                 + 2048)  # 4095
    decoded = decode_psf_flags(all_flags, return_bit_values=True)
    expected_all = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048]
    assert set(decoded) == set(expected_all)
    assert len(decoded) == 12

    # Test with array input
    flags_array = [0, 1, 2, 5]
    decoded_list = decode_psf_flags(flags_array, return_bit_values=True)
    assert len(decoded_list) == 4
    assert isinstance(decoded_list, list)

    # Check individual results
    assert decoded_list[0] == []
    assert decoded_list[1] == [1]
    assert decoded_list[2] == [2]
    assert set(decoded_list[3]) == {1, 4}

    # Test with numpy array
    flags_np = np.array([8, 16, 32])
    decoded_list = decode_psf_flags(flags_np, return_bit_values=True)
    assert len(decoded_list) == 3
    assert decoded_list[0] == [8]
    assert decoded_list[1] == [16]
    assert decoded_list[2] == [32]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for matching PSFs using Fourier methods.
"""

import numpy as np
from astropy.utils.decorators import deprecated
from scipy.fft import fft2, fftshift, ifft2

from photutils.psf_matching.utils import (_apply_window_to_fourier,
                                          _convert_psf_to_otf,
                                          _validate_kernel_inputs)

__all__ = ['create_matching_kernel', 'make_kernel', 'make_wiener_kernel']


def make_kernel(source_psf, target_psf, *, window=None, regularization=1e-4):
    """
    Make a convolution kernel that matches an input PSF to a target PSF
    using the ratio of Fourier transforms.

    This function computes the matching kernel in the Fourier domain by
    dividing the target PSF's Fourier transform by the source PSF's
    Fourier transform. To avoid division by near-zero values, the Fourier
    ratio is set to zero at frequencies where the source OTF (Optical
    Transfer Function) amplitude falls below a threshold.

    The kernel is computed as:

    .. math::

        K = \\mathcal{F}^{-1} \\left[ W \\cdot R \\right]

    where the Fourier-space ratio :math:`R` is defined as:

    .. math::

        R = \\begin{cases}
            \\frac{T}{S} & \\text{if } |S| > \\lambda \\cdot \\max(|S|) \\\\
            0 & \\text{otherwise}
            \\end{cases}

    Here, :math:`\\mathcal{F}^{-1}` is the inverse Fourier transform,
    :math:`S` and :math:`T` are the Fourier transforms of the source and
    target PSFs (the optical transfer functions, OTFs), :math:`\\lambda`
    is the ``regularization`` parameter, and :math:`W` is the optional
    ``window`` function (defaulting to 1 if not provided).

    Parameters
    ----------
    source_psf : 2D `~numpy.ndarray`
        The source PSF. The source PSF should have higher resolution
        (i.e., narrower) than the target PSF. ``source_psf`` and
        ``target_psf`` must have the same shape and pixel scale. It is
        assumed to be centered on the central pixel.

    target_psf : 2D `~numpy.ndarray`
        The target PSF. The target PSF should have lower resolution
        (i.e., broader) than the source PSF. ``source_psf`` and
        ``target_psf`` must have the same shape and pixel scale. It is
        assumed to be centered on the central pixel.

    window : callable, optional
        The window (taper) function or callable class instance used
        to remove high frequency noise from the PSF matching kernel.
        The window function should be a callable that accepts a single
        ``shape`` parameter (a tuple defining the 2D array shape) and
        returns a 2D array of the same shape. The returned window
        values must be in the range [0, 1], where 1.0 indicates full
        preservation of that spatial frequency and 0.0 indicates
        complete suppression. The window must be centered on the central
        pixel. Built-in window classes include:

        * `~photutils.psf_matching.HanningWindow`
        * `~photutils.psf_matching.TukeyWindow`
        * `~photutils.psf_matching.CosineBellWindow`
        * `~photutils.psf_matching.SplitCosineBellWindow`
        * `~photutils.psf_matching.TopHatWindow`

        For more information on window functions, custom windows, and
        example usage, see :ref:`PSF Matching <psf_matching>`.

    regularization : float, optional
        The regularization parameter that controls the OTF amplitude
        threshold for the source OTF (Optical Transfer Function, the
        Fourier transform of the PSF). At frequencies where the source
        OTF amplitude is below ``regularization`` times the peak
        amplitude, the Fourier ratio is set to zero to avoid division by
        near-zero values. Must be in the range [0, 1), where 0 provides
        no thresholding (only exact zeros are excluded) and values
        closer to 1 apply more aggressive thresholding. A value of 1
        would zero out all frequencies and produce a degenerate kernel.

    Returns
    -------
    kernel : 2D `~numpy.ndarray`
        The matching kernel to go from ``source_psf`` to ``target_psf``.
        The output matching kernel is normalized such that it sums to 1.

    Raises
    ------
    ValueError
        If the PSFs are not 2D arrays, have even dimensions, or do not
        have the same shape, if ``regularization`` is not in the range
        [0, 1), or if the window function output is invalid (not a 2D
        array, wrong shape, or values outside [0, 1]).

    TypeError
        If the input ``window`` is not callable.

    See Also
    --------
    make_wiener_kernel : Make a matching kernel using Wiener
                         (Tikhonov) regularization instead of
                         hard amplitude thresholding.

    Examples
    --------
    Make a matching kernel between two Gaussian PSFs:

    >>> import numpy as np
    >>> from astropy.modeling.models import Gaussian2D
    >>> from photutils.psf_matching import make_kernel
    >>> y, x = np.mgrid[0:51, 0:51]
    >>> psf1 = Gaussian2D(100, 25, 25, 3, 3)(x, y)
    >>> psf2 = Gaussian2D(100, 25, 25, 5, 5)(x, y)
    >>> psf1 /= psf1.sum()
    >>> psf2 /= psf2.sum()
    >>> kernel = make_kernel(psf1, psf2)
    >>> print(f'{kernel.sum():.1f}')
    1.0
    """
    source_psf, target_psf = _validate_kernel_inputs(
        source_psf, target_psf, window)

    if not 0 <= regularization < 1:
        msg = (f'regularization must be in the range [0, 1), '
               f'got {regularization}.')
        raise ValueError(msg)

    source_otf = fft2(source_psf)
    target_otf = fft2(target_psf)

    # Note: the following calculations are performed in the Fourier
    # domain with the DC component at the corner of the array (standard
    # FFT layout).

    # Regularized division to avoid dividing by near-zero values
    abs_source_otf = np.abs(source_otf)
    max_otf = np.max(abs_source_otf)
    mask = abs_source_otf > regularization * max_otf
    ratio = np.zeros_like(source_otf)  # dtype='complex128' from fft2
    ratio[mask] = target_otf[mask] / source_otf[mask]

    # Apply a window function in frequency space
    if window is not None:
        ratio = _apply_window_to_fourier(ratio, window, target_psf.shape)

    kernel = np.real(fftshift(ifft2(ratio)))
    if np.sum(kernel) < 1e-30:
        msg = ('The computed kernel sums to zero, which likely indicates '
               'that the regularization threshold is too high. Try reducing '
               'the regularization parameter or using a different window '
               'function.')
        raise ValueError(msg)

    return kernel / kernel.sum()


def make_wiener_kernel(source_psf, target_psf, *, regularization=1e-4,
                       penalty=None, window=None):
    """
    Make a convolution kernel that matches an input PSF to a target PSF
    using Wiener regularization in Fourier space.

    This function computes a Wiener-regularized PSF-matching kernel
    in the Fourier domain. The denominator includes a regularization
    term that stabilizes inversion of the source OTF (Optical Transfer
    Function, the Fourier transform of the PSF) by preventing division
    by small values, thereby suppressing noise amplification at spatial
    frequencies where the source response is weak.

    When no ``penalty`` is provided, the regularization is a
    frequency-independent (zero-order scalar) Tikhonov term expressed as
    a fraction of the peak power in the source OTF. In this case, the
    kernel is computed as:

    .. math::

        K = \\mathcal{F}^{-1} \\left[ W \\cdot
            \\frac{T \\cdot S^{*}}
                  {|S|^{2} + \\lambda \\cdot \\max(|S|^{2})} \\right]

    When a ``penalty`` array is provided (e.g., a Laplacian operator),
    the regularization becomes frequency-dependent:

    .. math::

        K = \\mathcal{F}^{-1} \\left[ W \\cdot
            \\frac{T \\cdot S^{*}}
                  {|S|^{2} + \\lambda \\cdot |P|^{2}} \\right]

    where :math:`P` is the OTF of the ``penalty`` operator. This
    penalizes high spatial frequencies more heavily, which is
    particularly effective at suppressing noise amplification.

    In both equations, :math:`\\mathcal{F}^{-1}` is the inverse Fourier
    transform, :math:`S` and :math:`T` are the Fourier transforms of
    the source and target PSFs (the OTFs), :math:`S^{*}` is the complex
    conjugate of :math:`S`, :math:`\\lambda` is the ``regularization``
    parameter, and :math:`W` is the optional ``window`` function
    (defaulting to 1 if not provided). :math:`|S|^{2}` is the power
    spectrum of the source OTF.

    When the ``penalty`` is set to ``'laplacian'``, the regularization
    reproduces the approach used by the ``pypher`` package (`Boucaud
    et al. 2016`_), which applies a discrete Laplacian operator as
    the penalty. This provides stronger suppression of high spatial
    frequencies, which can be beneficial when working with noisy or
    undersampled PSFs.

    Compared to `~photutils.psf_matching.make_kernel`, which uses a hard
    threshold on Fourier amplitude, this approach provides continuous,
    smooth regularization that is better suited for PSFs that have
    near-zero power at high spatial frequencies. The hard-cutoff
    approach zeros out frequencies where the source amplitude is
    below a threshold, which can introduce discontinuities in Fourier
    space. Wiener regularization instead smoothly down-weights those
    frequencies, producing matching kernels with less ringing.

    .. _Boucaud et al. 2016:
        https://ui.adsabs.harvard.edu/abs/2016A%26A...596A..63B/abstract

    Parameters
    ----------
    source_psf : 2D `~numpy.ndarray`
        The source PSF. The source PSF should have higher resolution
        (i.e., narrower) than the target PSF. ``source_psf`` and
        ``target_psf`` must have the same shape and pixel scale. It is
        assumed to be centered on the central pixel.

    target_psf : 2D `~numpy.ndarray`
        The target PSF. The target PSF should have lower resolution
        (i.e., broader) than the source PSF. ``source_psf`` and
        ``target_psf`` must have the same shape and pixel scale. It is
        assumed to be centered on the central pixel.

    regularization : float, optional
        The regularization parameter that controls the strength
        of the Wiener (Tikhonov) regularization. When ``penalty``
        is `None`, this is expressed as a fraction of the peak
        power in the source PSF's Fourier transform. When
        ``penalty`` is provided, this scales the penalty operator's
        power spectrum directly. Larger values produce smoother but
        less accurate matching kernels; smaller values preserve more
        detail but may amplify noise. Must be a positive number.

    penalty : `None`, ``'laplacian'``, ``'biharmonic'``, or 2D \
`~numpy.ndarray`, optional
        The regularization penalty operator. This controls the
        structure of the regularization term in the denominator:

        * `None` (default): Scalar Tikhonov regularization. The
          denominator is :math:`|S|^2 + \\lambda \\cdot \\max(|S|^2)`,
          providing uniform regularization across all spatial
          frequencies. Use this for well-behaved PSFs or when you want
          simple, frequency-independent smoothing.

        * ``'laplacian'``: Uses a discrete Laplacian operator (second
          derivative) as the penalty, producing frequency-dependent
          regularization that penalizes high spatial frequencies more
          heavily. The denominator becomes :math:`|S|^2 + \\lambda
          \\cdot |L|^2` where :math:`L` is the OTF of the Laplacian
          kernel ``[[0, -1, 0], [-1, 4, -1], [0, -1, 0]]``. This
          reproduces the regularization used by the ``pypher`` package
          (`Boucaud et al. 2016`_). This is the most commonly used
          penalty for PSF matching and works well for most applications.
          Requires PSFs to be at least 3x3.

        * ``'biharmonic'``: Uses a biharmonic operator (fourth
          derivative, Laplacian of the Laplacian) as the penalty,
          producing very strong suppression of high spatial frequencies.
          Uses the kernel ``[[0, 0, 1, 0, 0], [0, 2, -8, 2, 0], [1,
          -8, 20, -8, 1], [0, 2, -8, 2, 0], [0, 0, 1, 0, 0]]``. This
          produces the smoothest matching kernels and is useful when
          working with very noisy or poorly sampled PSFs, at the cost
          of reduced accuracy in matching. Requires PSFs to be at least
          5x5.

        * 2D `~numpy.ndarray`: A custom penalty operator array.
          Its OTF will be computed and used in the denominator as
          :math:`|S|^2 + \\lambda \\cdot |P|^2`. The PSFs must be at
          least as large as the penalty array in both dimensions.

    window : callable, optional
        The window (taper) function or callable class instance used
        to remove high frequency noise from the PSF matching kernel.
        The window function should be a callable that accepts a single
        ``shape`` parameter (a tuple defining the 2D array shape) and
        returns a 2D array of the same shape. The returned window
        values must be in the range [0, 1], where 1.0 indicates full
        preservation of that spatial frequency and 0.0 indicates
        complete suppression. The window must be centered on the central
        pixel. Built-in window classes include:

        * `~photutils.psf_matching.HanningWindow`
        * `~photutils.psf_matching.TukeyWindow`
        * `~photutils.psf_matching.CosineBellWindow`
        * `~photutils.psf_matching.SplitCosineBellWindow`
        * `~photutils.psf_matching.TopHatWindow`

        A window function is generally not needed when using Wiener
        regularization because the regularization itself suppresses
        high-frequency noise. However, a window may still be useful
        when working with noisy or undersampled PSFs.

        For more information on window functions, custom windows, and
        example usage, see :ref:`PSF Matching <psf_matching>`.

    Returns
    -------
    kernel : 2D `~numpy.ndarray`
        The matching kernel to go from ``source_psf`` to ``target_psf``.
        The output matching kernel is normalized such that it sums to 1.

    Raises
    ------
    ValueError
        If the PSFs are not 2D arrays, have even dimensions, do not have
        the same shape, are too small for the specified penalty, if
        ``regularization`` is not positive, or if ``penalty`` is not a
        valid value.

    TypeError
        If the input ``window`` is not callable.

    See Also
    --------
    make_kernel : Make a matching kernel using a hard frequency cutoff
                  instead of Wiener regularization.

    Examples
    --------
    Make a matching kernel between two Gaussian PSFs:

    >>> import numpy as np
    >>> from astropy.modeling.models import Gaussian2D
    >>> from photutils.psf_matching import make_wiener_kernel
    >>> y, x = np.mgrid[0:51, 0:51]
    >>> psf1 = Gaussian2D(100, 25, 25, 3, 3)(x, y)
    >>> psf2 = Gaussian2D(100, 25, 25, 5, 5)(x, y)
    >>> psf1 /= psf1.sum()
    >>> psf2 /= psf2.sum()
    >>> kernel = make_wiener_kernel(psf1, psf2)
    >>> print(f'{kernel.sum():.1f}')
    1.0

    Use the Laplacian penalty for frequency-dependent regularization:

    >>> kernel = make_wiener_kernel(psf1, psf2, penalty='laplacian')
    >>> print(f'{kernel.sum():.1f}')
    1.0

    Use the biharmonic penalty for maximum smoothness:

    >>> kernel = make_wiener_kernel(psf1, psf2, penalty='biharmonic')
    >>> print(f'{kernel.sum():.1f}')
    1.0
    """
    source_psf, target_psf = _validate_kernel_inputs(
        source_psf, target_psf, window)

    if regularization <= 0:
        msg = 'regularization must be a positive number.'
        raise ValueError(msg)

    # Validate and build the penalty term
    if penalty is None:
        penalty_array = None
    elif isinstance(penalty, str):
        if penalty == 'laplacian':
            penalty_array = np.array([[+0, -1, +0],
                                      [-1, +4, -1],
                                      [+0, -1, +0]], dtype=float)
        elif penalty == 'biharmonic':
            penalty_array = np.array([[+0, +0, +1, +0, +0],
                                      [+0, +2, -8, +2, +0],
                                      [+1, -8, 20, -8, +1],
                                      [+0, +2, -8, +2, +0],
                                      [+0, +0, +1, +0, +0]], dtype=float)
        else:
            msg = (f'Invalid penalty string {penalty!r}. '
                   'Must be "laplacian" or "biharmonic".')
            raise ValueError(msg)
    elif isinstance(penalty, np.ndarray):
        if penalty.ndim != 2:
            msg = 'penalty array must be 2D.'
            raise ValueError(msg)
        penalty_array = np.asarray(penalty, dtype=float)
    else:
        msg = ("penalty must be None, 'laplacian', 'biharmonic', or a 2D "
               'numpy array.')
        raise ValueError(msg)

    # Validate that PSF is large enough for the penalty
    if penalty_array is not None:
        penalty_shape = penalty_array.shape
        psf_shape = source_psf.shape
        if (psf_shape[0] < penalty_shape[0]
                or psf_shape[1] < penalty_shape[1]):
            msg = (f'PSFs must be at least as large as the penalty '
                   f'operator. PSF shape is {psf_shape}, but penalty '
                   f'shape is {penalty_shape}.')
            raise ValueError(msg)

    source_otf = fft2(source_psf)
    target_otf = fft2(target_psf)

    source_power = np.abs(source_otf) ** 2

    if penalty_array is not None:
        # Frequency-dependent regularization
        penalty_otf = _convert_psf_to_otf(penalty_array, source_psf.shape)
        reg_term = regularization * np.abs(penalty_otf) ** 2
    else:
        # Wiener (Tikhonov; scalar/zero-order) regularization.
        # This is frequency-independent and expressed as a fraction of
        # the peak power in the source OTF
        reg_term = regularization * np.max(source_power)

    # Compute the Wiener-regularized kernel in Fourier space
    kernel_otf = (target_otf * np.conj(source_otf)
                  / (source_power + reg_term))

    # Apply a window function in frequency space
    if window is not None:
        kernel_otf = _apply_window_to_fourier(
            kernel_otf, window, target_psf.shape)

    kernel = np.real(fftshift(ifft2(kernel_otf)))
    if np.sum(kernel) < 1e-30:
        msg = ('The computed kernel sums to zero, which likely indicates '
               'that the regularization threshold is too high. Try reducing '
               'the regularization parameter or using a different window '
               'function.')
        raise ValueError(msg)

    return kernel / kernel.sum()


@deprecated('3.0', alternative='make_kernel')
def create_matching_kernel(source_psf, target_psf, *, window=None,
                           regularization=1e-4):
    """
    Create a kernel to match 2D point spread functions (PSF).

    .. deprecated:: 3.0
        ``create_matching_kernel`` is deprecated as of Photutils 3.0 and
        will be removed in version 4.0. Use `make_kernel` instead.
    """
    return make_kernel(source_psf,
                       target_psf,
                       window=window,
                       regularization=regularization)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the circle module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from numpy.testing import assert_allclose

from photutils.aperture.circle import (CircularAnnulus, CircularAperture,
                                       SkyCircularAnnulus, SkyCircularAperture)
from photutils.aperture.tests.test_aperture_common import BaseTestAperture
from photutils.utils._optional_deps import HAS_MATPLOTLIB

POSITIONS = [(10, 20), (30, 40), (50, 60), (70, 80)]
RA, DEC = np.transpose(POSITIONS)
SKYCOORD = SkyCoord(ra=RA, dec=DEC, unit='deg')
UNIT = u.arcsec
RADII = (0.0, -1.0, -np.inf)


class TestCircularAperture(BaseTestAperture):
    aperture = CircularAperture(POSITIONS, r=3.0)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self):
        self.aperture.plot()

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_returns_patches(self):
        from matplotlib import pyplot as plt
        from matplotlib.patches import Patch

        my_patches = self.aperture.plot()
        assert isinstance(my_patches, list)
        for patch in my_patches:
            assert isinstance(patch, Patch)

        # Test creating a legend with these patches
        plt.legend(my_patches, list(range(len(my_patches))))

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'r' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            CircularAperture(POSITIONS, radius)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.r = 2.0
        assert aper != self.aperture


class TestCircularAnnulus(BaseTestAperture):
    aperture = CircularAnnulus(POSITIONS, r_in=3.0, r_out=7.0)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self):
        self.aperture.plot()

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_returns_patches(self):
        from matplotlib import pyplot as plt
        from matplotlib.patches import Patch

        my_patches = self.aperture.plot()
        assert isinstance(my_patches, list)

        for p in my_patches:
            assert isinstance(p, Patch)

        # Test creating a legend with these patches
        labels = list(range(len(my_patches)))
        _, ax = plt.subplots()
        ax.legend(my_patches, labels)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'r_in' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            CircularAnnulus(POSITIONS, r_in=radius, r_out=7.0)

        match = "'r_out' must be greater than 'r_in'"
        with pytest.raises(ValueError, match=match):
            CircularAnnulus(POSITIONS, r_in=3.0, r_out=radius)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.r_in = 2.0
        assert aper != self.aperture


class TestSkyCircularAperture(BaseTestAperture):
    aperture = SkyCircularAperture(SKYCOORD, r=3.0 * UNIT)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'r' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyCircularAperture(SKYCOORD, r=radius * UNIT)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.r = 2.0 * UNIT
        assert aper != self.aperture


class TestSkyCircularAnnulus(BaseTestAperture):
    aperture = SkyCircularAnnulus(SKYCOORD, r_in=3.0 * UNIT, r_out=7.0 * UNIT)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'r_in' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyCircularAnnulus(SKYCOORD, r_in=radius * UNIT, r_out=7.0 * UNIT)

        match = "'r_out' must be greater than 'r_in'"
        with pytest.raises(ValueError, match=match):
            SkyCircularAnnulus(SKYCOORD, r_in=3.0 * UNIT, r_out=radius * UNIT)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.r_in = 2.0 * UNIT
        assert aper != self.aperture

    @staticmethod
    def test_r_out_less_than_r_in():
        """
        Test that a ValueError is raised when r_out <= r_in.
        """
        match = "'r_out' must be greater than 'r_in'"
        with pytest.raises(ValueError, match=match):
            SkyCircularAnnulus(SKYCOORD, r_in=7.0 * UNIT, r_out=3.0 * UNIT)

    @staticmethod
    def test_non_angle_quantity():
        """
        Test that a ValueError is raised when r_in has non-angular
        units.
        """
        match = "'r_in' must have angular units"
        with pytest.raises(ValueError, match=match):
            SkyCircularAnnulus(SKYCOORD, r_in=0.5 * u.pix, r_out=7.0 * u.pix)


def test_slicing():
    xypos = [(10, 10), (20, 20), (30, 30)]
    aper1 = CircularAperture(xypos, r=3)
    aper2 = aper1[0:2]
    assert len(aper2) == 2

    aper3 = aper1[0]
    assert aper3.isscalar
    match = "A scalar 'CircularAperture' object has no len"
    with pytest.raises(TypeError, match=match):
        len(aper3)

    match = "A scalar 'CircularAperture' object cannot be indexed"
    with pytest.raises(TypeError, match=match):
        _ = aper3[0]


def test_area_overlap():
    data = np.ones((11, 11))
    xypos = [(0, 0), (5, 5), (50, 50)]
    aper = CircularAperture(xypos, r=3)
    areas = aper.area_overlap(data)
    assert_allclose(areas, [10.304636, np.pi * 9.0, np.nan])

    data2 = np.ones((11, 11)) * u.Jy
    areas = aper.area_overlap(data2)
    assert not isinstance(areas[0], u.Quantity)
    assert_allclose(areas, [10.304636, np.pi * 9.0, np.nan])

    aper2 = CircularAperture(xypos[1], r=3)
    area2 = aper2.area_overlap(data)
    assert_allclose(area2, np.pi * 9.0)

    area2 = aper2.area_overlap(data2)
    assert not isinstance(area2, u.Quantity)
    assert_allclose(area2, np.pi * 9.0)


def test_area_overlap_mask():
    data = np.ones((11, 11))
    mask = np.zeros((11, 11), dtype=bool)
    mask[0, 0:2] = True
    mask[5, 5:7] = True
    xypos = [(0, 0), (5, 5), (50, 50)]
    aper = CircularAperture(xypos, r=3)
    areas = aper.area_overlap(data, mask=mask)
    areas_exp = np.array([10.304636, np.pi * 9.0, np.nan]) - 2.0
    assert_allclose(areas, areas_exp)

    mask = np.zeros((3, 3), dtype=bool)
    match = 'mask and data must have the same shape'
    with pytest.raises(ValueError, match=match):
        aper.area_overlap(data, mask=mask)


def test_invalid_positions():
    match = r"'positions' must be a \(x, y\) pixel position or a list"
    with pytest.raises(ValueError, match=match):
        _ = CircularAperture([], r=3)

    with pytest.raises(ValueError, match=match):
        _ = CircularAperture([1], r=3)

    with pytest.raises(ValueError, match=match):
        _ = CircularAperture([[1]], r=3)

    with pytest.raises(ValueError, match=match):
        _ = CircularAperture([1, 2, 3], r=3)

    with pytest.raises(ValueError, match=match):
        _ = CircularAperture([[1, 2, 3]], r=3)

    x = np.arange(3)
    y = np.arange(3)
    xypos = np.transpose((x, y)) * u.pix
    match = "'positions' must not be a Quantity"
    with pytest.raises(TypeError, match=match):
        _ = CircularAperture(xypos, r=3)

    x = np.arange(3) * u.pix
    y = np.arange(3)
    xypos = zip(x, y, strict=True)
    with pytest.raises(TypeError, match=match):
        _ = CircularAperture(xypos, r=3)

    x = np.arange(3) * u.pix
    y = np.arange(3) * u.pix
    xypos = zip(x, y, strict=True)
    with pytest.raises(TypeError, match=match):
        _ = CircularAperture(xypos, r=3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for rounding numpy arrays.
"""

import numpy as np


def round_half_away(a):
    """
    Round a float or array of floats to the nearest integer, rounding
    half away from zero.

    Parameters
    ----------
    a : float or array_like
        The input float or array.

    Returns
    -------
    result : int, float, or array_like
        The rounded values. Finite inputs are returned as integers.
        Non-finite inputs (NaN or infinity) are returned as floats,
        preserving the NaN or infinity value.

    Notes
    -----
    NaN and infinity values are preserved in the output. Arrays
    containing any non-finite value are returned as float arrays;
    all-finite arrays are returned as integer arrays.
    """
    data = np.atleast_1d(np.asarray(a, dtype=float))
    rounded = np.where(data >= 0, np.floor(data + 0.5),
                       np.ceil(data - 0.5))

    if np.isscalar(a):
        val = rounded[0]
        if not np.isfinite(a):
            return val
        return int(val)

    if np.isfinite(data).all():
        return rounded.astype(int)

    return rounded



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the iterative module.
"""


import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.fitting import TRFLSQFitter
from astropy.modeling.models import Gaussian2D
from astropy.nddata import NDData, StdDevUncertainty
from astropy.table import QTable, Table
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from photutils.background import LocalBackground, MMMBackground
from photutils.datasets import make_model_image, make_noise_image
from photutils.detection import DAOStarFinder
from photutils.psf import (CircularGaussianPRF, IterativePSFPhotometry,
                           SourceGrouper, make_psf_model, make_psf_model_image)
from photutils.utils.exceptions import NoDetectionsWarning


@pytest.fixture(name='test_data')
def fixture_test_data():
    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    model_shape = (9, 9)
    n_sources = 10
    shape = (101, 101)
    data, true_params = make_psf_model_image(shape, psf_model, n_sources,
                                             model_shape=model_shape,
                                             flux=(500, 700),
                                             min_separation=10, seed=0)
    noise = make_noise_image(data.shape, mean=0, stddev=1, seed=0)
    data += noise
    error = np.abs(noise)

    return data, error, true_params


def make_mock_finder(x_col, y_col):
    def finder(data, *, mask=None):  # noqa: ARG001
        source_table = Table()
        source_table[x_col] = [25.1]
        source_table[y_col] = [24.9]
        return source_table
    return finder


FINDER_COLUMN_NAMES = [
    ('x', 'y'),
    ('x_init', 'y_init'),
    ('xcentroid', 'ycentroid'),
    ('x_centroid', 'y_centroid'),
    ('xpos', 'ypos'),
    ('x_peak', 'y_peak'),
    ('xcen', 'ycen'),
    ('x_fit', 'y_fit'),
    ('x_invalid', 'y_invalid'),
]


@pytest.mark.parametrize('mode', ['new', 'all'])
def test_iterative_psf_photometry_compound(mode):
    x_stddev = y_stddev = 1.7
    psf_func = Gaussian2D(amplitude=1, x_mean=0, y_mean=0, x_stddev=x_stddev,
                          y_stddev=y_stddev)
    psf_model = make_psf_model(psf_func, x_name='x_mean', y_name='y_mean')
    psf_model.x_stddev_2.fixed = False
    psf_model.y_stddev_2.fixed = False

    other_params = {psf_model.flux_name: (500, 700)}

    model_shape = (9, 9)
    n_sources = 10
    shape = (101, 101)
    data, true_params = make_psf_model_image(shape, psf_model, n_sources,
                                             model_shape=model_shape,
                                             **other_params,
                                             min_separation=10, seed=0)
    noise = make_noise_image(data.shape, mean=0, stddev=1, seed=0)
    data += noise
    error = np.abs(noise)

    init_params = QTable()
    init_params['x'] = [54, 29, 80]
    init_params['y'] = [8, 26, 29]
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 3.0)
    grouper = SourceGrouper(min_separation=2)

    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     grouper=grouper, aperture_radius=4,
                                     sub_shape=fit_shape,
                                     mode=mode, maxiters=2)
    phot = psfphot(data, error=error, init_params=init_params)
    assert isinstance(phot, QTable)
    assert len(phot) == len(true_params)

    cols = ('x_stddev_2', 'y_stddev_2')
    suffixes = ('_init', '_fit', '_err')
    colnames = [col + suffix for suffix in suffixes for col in cols]
    for colname in colnames:
        assert colname in phot.colnames

    # test model and residual images
    psf_shape = (9, 9)
    model1 = psfphot.make_model_image(data.shape, psf_shape=psf_shape,
                                      include_local_bkg=False)
    resid1 = psfphot.make_residual_image(data, psf_shape=psf_shape,
                                         include_local_bkg=False)
    model2 = psfphot.make_model_image(data.shape, psf_shape=psf_shape,
                                      include_local_bkg=True)
    resid2 = psfphot.make_residual_image(data, psf_shape=psf_shape,
                                         include_local_bkg=True)
    assert model1.shape == data.shape
    assert model2.shape == data.shape
    assert resid1.shape == data.shape
    assert resid2.shape == data.shape
    assert_equal(data - model1, resid1)
    assert_equal(data - model2, resid2)

    # test with init_params
    init_params = psfphot.fit_results[-1].results_to_init_params()
    phot = psfphot(data, error=error, init_params=init_params)
    assert isinstance(phot, QTable)
    assert len(phot) == len(true_params)

    cols = ('x_stddev_2', 'y_stddev_2')
    suffixes = ('_init', '_fit', '_err')
    colnames = [col + suffix for suffix in suffixes for col in cols]
    for colname in colnames:
        assert colname in phot.colnames


def test_iterative_psf_photometry_mode_new(test_data):
    data, error, sources = test_data

    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    bkgstat = MMMBackground()
    local_bkg_estimator = LocalBackground(5, 10, bkg_estimator=bkgstat)
    finder = DAOStarFinder(10.0, 2.0)

    init_params = QTable()
    init_params['x'] = [54, 29, 80]
    init_params['y'] = [8, 26, 29]
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     mode='new',
                                     local_bkg_estimator=local_bkg_estimator,
                                     aperture_radius=4)
    phot = psfphot(data, error=error, init_params=init_params)
    cols = ['id', 'group_id', 'group_size', 'iter_detected', 'local_bkg']
    assert phot.colnames[:5] == cols
    assert len(psfphot.fit_results) == 2

    assert 'iter_detected' in phot.colnames
    assert len(phot) == len(sources)

    resid_data = psfphot.make_residual_image(data, psf_shape=fit_shape)
    assert isinstance(resid_data, np.ndarray)
    assert resid_data.shape == data.shape

    nddata = NDData(data)
    resid_nddata = psfphot.make_residual_image(nddata, psf_shape=fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert resid_nddata.data.shape == data.shape

    # test that repeated calls reset the results
    phot = psfphot(data, error=error, init_params=init_params)
    assert len(psfphot.fit_results) == 2

    # test NDData without units
    uncertainty = StdDevUncertainty(error)
    nddata = NDData(data, uncertainty=uncertainty)
    phot0 = psfphot(nddata, init_params=init_params)
    colnames = ('flux_init', 'flux_fit', 'flux_err', 'local_bkg')
    for col in colnames:
        assert_allclose(phot0[col], phot[col])
    resid_nddata = psfphot.make_residual_image(nddata, psf_shape=fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert_equal(resid_nddata.data, resid_data)

    # test with units and mode='new'
    unit = u.Jy
    finder_units = DAOStarFinder(10.0 * unit, 2.0)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape,
                                     finder=finder_units, mode='new',
                                     local_bkg_estimator=local_bkg_estimator,
                                     aperture_radius=4)

    phot2 = psfphot(data << unit, error=error << unit, init_params=init_params)
    assert phot2['flux_fit'].unit == unit
    colnames = ('flux_init', 'flux_fit', 'flux_err', 'local_bkg')
    for col in colnames:
        assert phot2[col].unit == unit
        assert_allclose(phot2[col].value, phot[col])

    # test NDData with units
    uncertainty = StdDevUncertainty(error << unit)
    nddata = NDData(data << unit, uncertainty=uncertainty)
    phot3 = psfphot(nddata, init_params=init_params)
    colnames = ('flux_init', 'flux_fit', 'flux_err', 'local_bkg')
    for col in colnames:
        assert phot3[col].unit == unit
        assert_allclose(phot3[col].value, phot2[col].value)
    resid_nddata = psfphot.make_residual_image(nddata, psf_shape=fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert resid_nddata.unit == unit

    # test return None if no stars are found on first iteration
    finder = DAOStarFinder(1000.0, 2.0)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     mode='new',
                                     local_bkg_estimator=local_bkg_estimator,
                                     aperture_radius=4)
    match = 'No sources were found'
    with pytest.warns(NoDetectionsWarning, match=match):
        phot = psfphot(data, error=error)
    assert phot is None


def test_iterative_psf_photometry_mode_all():
    sources = QTable()
    sources['x_0'] = [50, 45, 55, 27, 22, 77, 82]
    sources['y_0'] = [50, 52, 48, 27, 30, 77, 79]
    sources['flux'] = [1000, 100, 50, 1000, 100, 1000, 100]

    shape = (101, 101)
    psf_model = CircularGaussianPRF(flux=500, fwhm=9.4)
    psf_shape = (41, 41)
    data = make_model_image(shape, psf_model, sources, model_shape=psf_shape)

    fit_shape = (5, 5)
    finder = DAOStarFinder(0.2, fwhm=6.0, min_separation=0)
    sub_shape = psf_shape
    grouper = SourceGrouper(10)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     grouper=grouper, aperture_radius=4,
                                     sub_shape=sub_shape, mode='all',
                                     maxiters=3)
    phot = psfphot(data)
    cols = ['id', 'group_id', 'group_size', 'iter_detected', 'local_bkg']
    assert phot.colnames[:5] == cols

    assert len(phot) == 7
    assert_equal(phot['group_id'], [1, 2, 3, 1, 2, 2, 3])
    assert_equal(phot['iter_detected'], [1, 1, 1, 2, 2, 2, 2])
    assert_allclose(phot['flux_fit'], [1000, 1000, 1000, 100, 50, 100, 100])

    resid = psfphot.make_residual_image(data, psf_shape=sub_shape)
    assert_allclose(resid, 0, atol=1e-6)

    match = "mode must be 'new' or 'all'"
    with pytest.raises(ValueError, match=match):
        psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                         grouper=grouper, aperture_radius=4,
                                         sub_shape=sub_shape, mode='invalid')

    match = "grouper must be input for the 'all' mode"
    with pytest.raises(ValueError, match=match):
        psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                         grouper=None, aperture_radius=4,
                                         sub_shape=sub_shape, mode='all')

    # test with units and mode='all'
    unit = u.Jy
    finderu = DAOStarFinder(0.2 * unit, fwhm=6.0, min_separation=0)
    psfphotu = IterativePSFPhotometry(psf_model, fit_shape, finder=finderu,
                                      grouper=grouper, aperture_radius=4,
                                      sub_shape=sub_shape, mode='all',
                                      maxiters=3)
    phot2 = psfphotu(data << unit)
    assert len(phot2) == 7
    assert_equal(phot2['group_id'], [1, 2, 3, 1, 2, 2, 3])
    assert_equal(phot2['iter_detected'], [1, 1, 1, 2, 2, 2, 2])
    colnames = ('flux_init', 'flux_fit', 'flux_err', 'local_bkg')
    for col in colnames:
        assert phot2[col].unit == unit
        assert_allclose(phot2[col].value, phot[col])

    # test NDData with units
    nddata = NDData(data * unit)
    phot3 = psfphotu(nddata)
    colnames = ('flux_init', 'flux_fit', 'flux_err', 'local_bkg')
    for col in colnames:
        assert phot3[col].unit == unit
        assert_allclose(phot3[col].value, phot[col])
    resid_nddata = psfphotu.make_residual_image(nddata, psf_shape=fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert resid_nddata.unit == unit


def test_iterative_methods(test_data):
    data, error, sources = test_data

    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    finder = DAOStarFinder(10.0, 2.0)

    init_params = QTable()
    init_params['x'] = [54, 29, 80]
    init_params['y'] = [8, 26, 29]
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     mode='new',
                                     aperture_radius=4)

    match = 'No results available. Please run the IterativePSFPhotometry'
    with pytest.raises(ValueError, match=match):
        psfphot.make_model_image(data.shape)
    with pytest.raises(ValueError, match=match):
        psfphot.make_residual_image(data)

    phot = psfphot(data, error=error, init_params=init_params)
    cols = ['id', 'group_id', 'group_size', 'iter_detected', 'local_bkg']
    assert phot.colnames[:5] == cols
    assert len(psfphot.fit_results) == 2

    init_params = psfphot.results_to_init_params()
    assert isinstance(init_params, QTable)
    assert len(init_params) == len(sources)

    model_params = psfphot.results_to_model_params()
    assert isinstance(model_params, QTable)
    assert len(model_params) == len(sources)


def test_iterative_psf_photometry_overlap():
    """
    Regression test for #1769.

    A ValueError should not be raised for no overlap.
    """
    fwhm = 3.5
    psf_model = CircularGaussianPRF(flux=1, fwhm=fwhm)
    data, _ = make_psf_model_image((150, 150), psf_model, n_sources=300,
                                   model_shape=(11, 11), flux=(50, 100),
                                   min_separation=1, seed=0)
    noise = make_noise_image(data.shape, mean=0, stddev=0.01, seed=0)
    data += noise
    error = np.abs(noise)
    slc = (slice(0, 50), slice(0, 50))
    data = data[slc]
    error = error[slc]

    daofinder = DAOStarFinder(threshold=0.5, fwhm=fwhm, min_separation=0)
    grouper = SourceGrouper(min_separation=1.3 * fwhm)
    fitter = TRFLSQFitter()
    fit_shape = (5, 5)
    sub_shape = fit_shape
    psfphot = IterativePSFPhotometry(psf_model, fit_shape=fit_shape,
                                     finder=daofinder, mode='all',
                                     grouper=grouper, maxiters=2,
                                     sub_shape=sub_shape,
                                     aperture_radius=3, fitter=fitter)
    match = r'One or more .* may not have converged'
    with pytest.warns(AstropyUserWarning, match=match):
        phot = psfphot(data, error=error)
    assert len(phot) == 38


def test_iterative_psf_photometry_subshape():
    """
    A ValueError should not be raised if sub_shape=None and the model
    does not have a bounding box.
    """
    fwhm = 3.5
    psf_model = CircularGaussianPRF(flux=1, fwhm=fwhm)
    data, _ = make_psf_model_image((150, 150), psf_model, n_sources=30,
                                   model_shape=(11, 11), flux=(50, 100),
                                   min_separation=1, seed=0)

    daofinder = DAOStarFinder(threshold=0.5, fwhm=fwhm)
    grouper = SourceGrouper(min_separation=1.3 * fwhm)
    fitter = TRFLSQFitter()
    fit_shape = (5, 5)
    sub_shape = None
    psf_model.bounding_box = None
    psfphot = IterativePSFPhotometry(psf_model, fit_shape=fit_shape,
                                     finder=daofinder, mode='all',
                                     grouper=grouper, maxiters=2,
                                     sub_shape=sub_shape,
                                     aperture_radius=3, fitter=fitter)
    match = r'model_shape must be specified .* does not have a bounding_box'
    with pytest.raises(ValueError, match=match):
        psfphot(data)


def test_iterative_psf_photometry_inputs():
    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    finder = DAOStarFinder(10.0, 2.0)

    match = 'finder cannot be None for IterativePSFPhotometry'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=None,
                                   aperture_radius=4)

    match = 'aperture_radius cannot be None for IterativePSFPhotometry'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=None)

    match = 'maxiters must be a strictly-positive scalar'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=-1)
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=[1, 2])

    match = 'maxiters must be an integer'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=3.14)


@pytest.mark.parametrize(('x_col', 'y_col'), FINDER_COLUMN_NAMES)
def test_iterative_finder_column_names(x_col, y_col):
    """
    Test that IterativePSFPhotometry works correctly with a finder that
    returns different column names for source positions.
    """
    finder = make_mock_finder(x_col, y_col)

    sources = Table()
    sources['id'] = [1]
    sources['flux'] = 7.0
    sources['x_0'] = 25.0
    sources['y_0'] = 25.0
    shape = (31, 31)
    psf_model = CircularGaussianPRF(flux=1.0, fwhm=3.1)
    data = make_model_image(shape, psf_model, sources)

    fit_shape = (9, 9)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     aperture_radius=10, maxiters=3)

    # invalid column names should raise an error
    if x_col == 'x_invalid' or y_col == 'y_invalid':
        match = 'must contain columns for x and y coordinates'
        with pytest.raises(ValueError, match=match):
            psfphot(data, init_params=sources)
        return

    phot_tbl = psfphot(data)

    assert_allclose(phot_tbl['x_init'][0], 25.1)
    assert_allclose(phot_tbl['y_init'][0], 24.9)
    assert_allclose(phot_tbl['x_fit'][0], 25.0, atol=1e-6)
    assert_allclose(phot_tbl['y_fit'][0], 25.0, atol=1e-6)
    assert_allclose(phot_tbl['flux_fit'][0], 7.0, rtol=1e-6)


def test_repr():
    psf_model = CircularGaussianPRF(flux=1.0, fwhm=3.1)
    fit_shape = (9, 9)
    finder = DAOStarFinder(6.0, 2.0)

    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     aperture_radius=10)
    cls_repr = repr(psfphot)
    assert cls_repr.startswith(f'{psfphot.__class__.__name__}(')


def test_move_column():
    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     aperture_radius=4, maxiters=1)
    tbl = QTable()
    tbl['a'] = [1, 2, 3]
    tbl['b'] = [4, 5, 6]
    tbl['c'] = [7, 8, 9]

    tbl1 = psfphot._move_column(tbl, 'a', 'c')
    assert tbl1.colnames == ['b', 'c', 'a']
    tbl2 = psfphot._move_column(tbl, 'd', 'b')
    assert tbl2.colnames == ['a', 'b', 'c']
    tbl3 = psfphot._move_column(tbl, 'b', 'b')
    assert tbl3.colnames == ['a', 'b', 'c']


def test_iterative_model_residual_image_nonfinite_localbkg(test_data):
    """
    Test that make_model_image and make_residual_image handle non-finite
    local background values correctly for IterativePSFPhotometry.

    When include_local_bkg=True and the local_bkg is non-finite (NaN or
    inf), the non-finite value should be treated as 0 and not included
    in the model or residual images.
    """
    data, error, _ = test_data

    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    finder = DAOStarFinder(10.0, 2.0)

    # Find sources and manually set some local_bkg values to non-finite
    sources = finder(data)
    sources['local_bkg'] = np.zeros(len(sources))
    sources['local_bkg'][0] = np.nan
    sources['local_bkg'][1] = np.inf
    if len(sources) > 2:
        sources['local_bkg'][2] = -np.inf

    # Perform iterative PSF photometry with non-finite local_bkg
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     aperture_radius=4, maxiters=2)
    phot = psfphot(data, error=error, init_params=sources)

    # Test make_model_image with include_local_bkg=True
    psf_shape = (15, 15)
    model_with_bkg = psfphot.make_model_image(data.shape, psf_shape=psf_shape,
                                              include_local_bkg=True)
    model_without_bkg = psfphot.make_model_image(
        data.shape, psf_shape=psf_shape, include_local_bkg=False)

    # The model images should be finite everywhere
    assert np.all(np.isfinite(model_with_bkg))
    assert np.all(np.isfinite(model_without_bkg))

    # For sources with non-finite local_bkg, the model with and without
    # local_bkg should be identical (since non-finite is treated as 0)
    # Check this by comparing the models at source positions
    for i in range(min(3, len(phot))):
        if not np.isfinite(phot['local_bkg'][i]):
            x_fit = int(phot['x_fit'][i])
            y_fit = int(phot['y_fit'][i])
            assert_allclose(model_with_bkg[y_fit, x_fit],
                            model_without_bkg[y_fit, x_fit], rtol=1e-6)


def test_iterative_residual_image_localbkg_invalid_sources(test_data):
    """
    Test that make_residual_image handles sources with non-finite
    local_bkg values and sources outside the image (invalid sources)
    correctly for IterativePSFPhotometry.
    """
    data, error, _ = test_data

    psf_model = CircularGaussianPRF(flux=1, fwhm=2.7)
    fit_shape = (5, 5)
    finder = DAOStarFinder(10.0, 2.0)

    sources = finder(data)

    # Add non-finite local_bkg values to init_params
    sources['local_bkg'] = np.zeros(len(sources))
    sources['local_bkg'][0] = np.nan
    sources['local_bkg'][1] = np.inf
    sources['local_bkg'][2] = -np.inf
    # Add an invalid source outside the image
    sources['x_centroid'][-3] = 1000
    sources['y_centroid'][-3] = 1000

    # Perform iterative PSF photometry with init_params containing
    # non-finite local_bkg
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     aperture_radius=4, maxiters=2)
    psfphot(data, error=error, init_params=sources)

    residual_img = psfphot.make_residual_image(data, include_local_bkg=True)

    assert residual_img.shape == data.shape
    assert np.all(np.isfinite(residual_img))


def test_decode_flags():
    """
    Test the decode_flags convenience method.
    """
    # Create test data with some sources that will have flags
    yy, xx = np.mgrid[:21, :21]
    psf_model = CircularGaussianPRF(flux=1, x_0=10, y_0=10, fwhm=2)

    # Source 1: normal source (no flags expected)
    m1 = CircularGaussianPRF(flux=100, x_0=10, y_0=10, fwhm=2)
    # Source 2: negative flux (will have negative_flux flag)
    m2 = CircularGaussianPRF(flux=-50, x_0=5, y_0=5, fwhm=2)
    # Source 3: outside bounds (will have outside_bounds flag)
    m3 = CircularGaussianPRF(flux=100, x_0=25, y_0=25, fwhm=2)

    data = m1(xx, yy) + m2(xx, yy) + m3(xx, yy)

    init_params = Table({
        'x': [10, 5, 25],
        'y': [10, 5, 25],
        'flux': [100, 100, 100],
    })

    finder = DAOStarFinder(6.0, 2.0)
    psfphot = IterativePSFPhotometry(psf_model, (3, 3), finder=finder,
                                     aperture_radius=4, maxiters=1)

    # Test that decode_flags raises ValueError before running photometry
    match = 'No results available'
    with pytest.raises(ValueError, match=match):
        psfphot.decode_flags()

    # Run photometry
    results = psfphot(data, init_params=init_params)

    # Test decode_flags method
    decoded_flags = psfphot.decode_flags()

    # Check that we get a list of lists
    assert isinstance(decoded_flags, list)
    assert len(decoded_flags) == len(results)

    # Each element should be a list of strings
    for decoded in decoded_flags:
        assert isinstance(decoded, list)
        for flag_name in decoded:
            assert isinstance(flag_name, str)

    # Check that the first source has no flags or minimal flags
    # (depending on fitting success)
    assert isinstance(decoded_flags[0], list)

    # Check that the second source has the negative_flux flag
    assert 'negative_flux' in decoded_flags[1]

    # Check that the third source has flags (it's outside the image bounds)
    # It should have 'no_overlap' since it's completely outside
    assert len(decoded_flags[2]) > 0
    assert 'no_overlap' in decoded_flags[2]

    # Verify that decode_flags gives the same result as calling
    # decode_psf_flags directly
    from photutils.psf.flags import decode_psf_flags
    direct_decoded = decode_psf_flags(results['flags'])
    assert decoded_flags == direct_decoded



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the interpolators module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.background.background_2d import Background2D
from photutils.background.interpolators import (BkgIDWInterpolator,
                                                _BkgZoomInterpolator)


@pytest.fixture
def test_data():
    """
    Create test data for interpolator tests.
    """
    return np.ones((300, 300))


@pytest.fixture
def test_mesh():
    """
    Create test mesh for interpolator tests.
    """
    return np.array([[0.01, 0.01, 0.02],
                     [0.01, 0.02, 0.03],
                     [0.03, 0.03, 12.9]])


def test_zoom_interp_constant_mesh(test_data):
    """
    Test the zoom interpolator with a constant-valued mesh.

    When all mesh values are equal, the interpolator takes an early-exit
    path that fills the output with the constant value directly,
    bypassing `scipy.ndimage.zoom` entirely. This path must produce the
    correct fill value both for plain arrays and for Quantity inputs.
    """
    bkg = Background2D(test_data, 100)
    interp = _BkgZoomInterpolator()

    constant_mesh = np.full((3, 3), 7.5)
    result = interp(constant_mesh, **bkg._interp_kwargs)
    assert result.shape == bkg._interp_kwargs['shape']
    assert np.all(result == 7.5)

    # Also verify with a Quantity mesh
    unit = u.nJy
    bkg_q = Background2D(test_data << unit, 100)
    result_q = interp(constant_mesh << unit, **bkg_q._interp_kwargs)
    assert result_q.shape == bkg_q._interp_kwargs['shape']
    assert np.all(result_q == 7.5)


def test_zoom_interp(test_data, test_mesh):
    """
    Test the zoom interpolator.
    """
    bkg = Background2D(test_data, 100)

    interp = _BkgZoomInterpolator(clip=False)
    zoom = interp(test_mesh, **bkg._interp_kwargs)
    assert zoom.shape == (300, 300)

    # Test with units
    unit = u.nJy
    bkg = Background2D(test_data << unit, 100)
    interp = _BkgZoomInterpolator(clip=False)
    zoom = interp(test_mesh << unit, **bkg._interp_kwargs)
    assert zoom.shape == (300, 300)

    # Test repr
    cls_repr = repr(interp)
    assert cls_repr.startswith(f'{interp.__class__.__name__}')


def test_zoom_interp_clip(test_data, test_mesh):
    """
    Test the zoom interpolator with clipping.
    """
    bkg = Background2D(test_data, 100)

    interp1 = _BkgZoomInterpolator(clip=False)
    zoom1 = interp1(test_mesh, **bkg._interp_kwargs)

    interp2 = _BkgZoomInterpolator(clip=True)
    zoom2 = interp2(test_mesh, **bkg._interp_kwargs)

    minval = np.min(test_mesh)
    maxval = np.max(test_mesh)
    assert np.min(zoom1) < minval
    assert np.max(zoom1) > maxval
    assert np.min(zoom2) == minval
    assert np.max(zoom2) == maxval


def test_idw_interp(test_data, test_mesh):
    """
    Test the IDW interpolator.
    """
    with pytest.warns(AstropyDeprecationWarning):
        interp = BkgIDWInterpolator()
    with pytest.warns(AstropyDeprecationWarning):
        bkg = Background2D(test_data, 100, interpolator=interp)

    zoom = interp(test_mesh, **bkg._interp_kwargs)
    assert zoom.shape == (300, 300)

    # Test constant mesh data
    zoom = interp(np.ones_like(test_mesh), **bkg._interp_kwargs)
    assert np.all(zoom == 1)

    # Test with units
    unit = u.nJy
    with pytest.warns(AstropyDeprecationWarning):
        bkg = Background2D(test_data << unit, 100, interpolator=interp)
    zoom = interp(test_mesh << unit, **bkg._interp_kwargs)
    assert zoom.shape == (300, 300)

    # Test repr
    cls_repr = repr(interp)
    assert cls_repr.startswith(f'{interp.__class__.__name__}')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.isophote.ellipse import Ellipse
from photutils.isophote.geometry import EllipseGeometry
from photutils.isophote.model import build_ellipse_model
from photutils.isophote.sample import CentralEllipseSample, EllipseSample
from photutils.isophote.tests.make_test_data import make_test_image


class TestEllipsePositionalKwargs:
    """
    Test Ellipse.__init__, fit_image, fit_isophote for positional
    optional args.
    """

    def setup_method(self):
        self.data = make_test_image(seed=0)

    def test_init_positional_warns(self):
        geometry = EllipseGeometry(x0=256, y0=256, sma=10, eps=0.2,
                                   pa=np.pi / 2)
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            Ellipse(self.data, geometry)

    def test_init_keyword_no_warning(self):
        geometry = EllipseGeometry(x0=256, y0=256, sma=10, eps=0.2,
                                   pa=np.pi / 2)
        Ellipse(self.data, geometry=geometry)

    def test_fit_image_positional_warns(self):
        ellipse = Ellipse(self.data)
        match = 'fit_image'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            ellipse.fit_image(10.0)

    def test_fit_image_keyword_no_warning(self):
        ellipse = Ellipse(self.data)
        ellipse.fit_image(sma0=10.0)

    def test_fit_isophote_positional_warns(self):
        ellipse = Ellipse(self.data)
        match = 'fit_isophote'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            ellipse.fit_isophote(40.0, 0.1)

    def test_fit_isophote_keyword_no_warning(self):
        ellipse = Ellipse(self.data)
        ellipse.fit_isophote(40.0, step=0.1)


class TestEllipseGeometryPositionalKwargs:
    """
    Test EllipseGeometry.__init__ and find_center.
    """

    def test_init_positional_warns(self):
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            EllipseGeometry(0.0, 0.0, 100.0, 0.0, 0.0, 0.2)

    def test_init_keyword_no_warning(self):
        EllipseGeometry(0.0, 0.0, 100.0, 0.0, 0.0, astep=0.2)

    def test_find_center_positional_warns(self):
        data = make_test_image(seed=0)
        geometry = EllipseGeometry(x0=256, y0=256, sma=10, eps=0.2,
                                   pa=np.pi / 2)
        match = 'find_center'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            geometry.find_center(data, 0.5)

    def test_find_center_keyword_no_warning(self):
        data = make_test_image(seed=0)
        geometry = EllipseGeometry(x0=256, y0=256, sma=10, eps=0.2,
                                   pa=np.pi / 2)
        geometry.find_center(data, threshold=0.5)


class TestBuildEllipseModelPositionalKwargs:
    """
    Test build_ellipse_model.
    """

    def test_positional_warns(self):
        data = make_test_image(seed=0)
        ellipse = Ellipse(data)
        isolist = ellipse.fit_image(sma0=10.0)
        match = 'build_ellipse_model'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            build_ellipse_model(data.shape, isolist, 0.0)

    def test_keyword_no_warning(self):
        data = make_test_image(seed=0)
        ellipse = Ellipse(data)
        isolist = ellipse.fit_image(sma0=10.0)
        build_ellipse_model(data.shape, isolist, fill=0.0)


class TestEllipseSamplePositionalKwargs:
    """
    Test EllipseSample.__init__ and update.
    """

    def setup_method(self):
        self.data = make_test_image(seed=0)

    def test_init_positional_warns(self):
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            EllipseSample(self.data, 40.0, 256.0)

    def test_init_keyword_no_warning(self):
        EllipseSample(self.data, 40.0, x0=256.0)

    def test_update_positional_warns(self):
        sample = EllipseSample(self.data, 40.0)
        fix = np.array([False, False, False, False])
        match = 'update'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            sample.update(fix)

    def test_update_keyword_no_warning(self):
        sample = EllipseSample(self.data, 40.0)
        fix = np.array([False, False, False, False])
        sample.update(fixed_parameters=fix)


class TestCentralEllipseSamplePositionalKwargs:
    """
    Test CentralEllipseSample.update.
    """

    def test_update_positional_warns(self):
        data = make_test_image(seed=0)
        sample = CentralEllipseSample(data, 0.0)
        fix = np.array([False, False, False, False])
        match = 'update'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            sample.update(fix)

    def test_update_keyword_no_warning(self):
        data = make_test_image(seed=0)
        sample = CentralEllipseSample(data, 0.0)
        fix = np.array([False, False, False, False])
        sample.update(fixed_parameters=fix)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the noise module.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from photutils.datasets import apply_poisson_noise, make_noise_image


def test_apply_poisson_noise():
    """
    Test if Poisson noise is applied correctly.
    """
    shape = (100, 100)
    data = np.ones(shape)
    result = apply_poisson_noise(data)
    assert result.shape == shape
    assert_allclose(result.mean(), 1.0, atol=1.0)


def test_apply_poisson_noise_dtype():
    """
    Test if the output data type is the same as the input data type.
    """
    data = np.ones((10, 10), dtype=float) * 5.0
    result = apply_poisson_noise(data, seed=0)
    assert result.dtype == data.dtype


def test_apply_poisson_noise_seed():
    """
    Test if Poisson noise is reproducible with a given seed.
    """
    shape = (100, 100)
    data = np.ones(shape) * 10.0
    result1 = apply_poisson_noise(data, seed=12345)
    result2 = apply_poisson_noise(data, seed=12345)
    assert_allclose(result1, result2)


def test_apply_poisson_noise_negative():
    """
    Test if negative image values raises ValueError.
    """
    shape = (100, 100)
    data = np.zeros(shape) - 1.0
    match = 'data must not contain any negative values'
    with pytest.raises(ValueError, match=match):
        apply_poisson_noise(data)


def test_make_noise_image():
    """
    Test if noise image is generated correctly.
    """
    shape = (100, 100)
    image = make_noise_image(shape, distribution='gaussian', mean=0.0,
                             stddev=2.0)
    assert image.shape == shape
    assert_allclose(image.mean(), 0.0, atol=1.0)


def test_make_noise_image_poisson():
    """
    Test noise image with Poisson noise.
    """
    shape = (100, 100)
    image = make_noise_image(shape, distribution='poisson', mean=1.0)
    assert image.shape == shape
    assert_allclose(image.mean(), 1.0, atol=1.0)


def test_make_noise_image_nomean():
    """
    Test invalid inputs to make_noise_image.
    """
    shape = (100, 100)
    match = 'Invalid distribution:'
    with pytest.raises(ValueError, match=match):
        make_noise_image(shape, distribution='invalid', mean=0, stddev=2.0)

    match = "'mean' must be input"
    with pytest.raises(ValueError, match=match):
        make_noise_image(shape, distribution='gaussian', stddev=2.0)

    match = "'stddev' must be input for Gaussian noise"
    with pytest.raises(ValueError, match=match):
        make_noise_image(shape, distribution='gaussian', mean=2.0)


def test_make_noise_image_seed():
    """
    Test if noise image is reproducible with a given seed.
    """
    shape = (100, 100)
    image1 = make_noise_image(shape, distribution='gaussian', mean=0.0,
                              stddev=2.0, seed=12345)
    image2 = make_noise_image(shape, distribution='gaussian', mean=0.0,
                              stddev=2.0, seed=12345)
    assert_allclose(image1, image2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for interpolating data.
"""

import numpy as np
from scipy.spatial import cKDTree

from photutils.utils._deprecation import (deprecated_positional_kwargs,
                                          deprecated_renamed_argument)

__all__ = ['ShepardIDWInterpolator']


class ShepardIDWInterpolator:
    """
    Class to perform Inverse Distance Weighted (IDW) interpolation.

    This interpolator uses a modified version of `Shepard's method
    <https://en.wikipedia.org/wiki/Inverse_distance_weighting>`_ (see
    the Notes section for details).

    Parameters
    ----------
    coordinates : float, 1D array_like, or NxM array_like
        Coordinates of the known data points. In general, it is expected
        that these coordinates are in a form of an NxM-like array where
        N is the number of points and M is dimension of the coordinate
        space. When M=1 (1D space), then the ``coordinates`` parameter
        may be entered as a 1D array or, if only one data point is
        available, ``coordinates`` can be a scalar number representing
        the 1D coordinate of the data point.

        .. note::
            If the dimensionality of ``coordinates`` is larger than 2,
            e.g., if it is of the form N1 x N2 x N3 x ... x Nn x M, then
            it will be flattened to form an array of size NxM where N =
            N1 * N2 * ... * Nn.

    values : float or 1D array_like
        Values of the data points corresponding to each coordinate
        provided in ``coordinates``. In general a 1D array is expected.
        When a single data point is available, then ``values`` can be a
        scalar number.

        .. note::
            If the dimensionality of ``values`` is larger than 1 then it
            will be flattened.

    weights : float or 1D array_like, optional
        Weights to be associated with each data value. These weights, if
        provided, will be combined with inverse distance weights (see
        the Notes section for details). When ``weights`` is `None`
        (default), then only inverse distance weights will be used. When
        provided, this input parameter must have the same form as
        ``values``.

    leafsize : float, optional
        The number of points at which the k-d tree algorithm switches
        over to brute-force. ``leafsize`` must be positive. See
        `scipy.spatial.cKDTree` for further information.

    Notes
    -----
    This interpolator uses a slightly modified version of `Shepard's
    method <https://en.wikipedia.org/wiki/Inverse_distance_weighting>`_.
    The essential difference is the introduction of a "regularization"
    parameter (``reg``) that is used when computing the inverse distance
    weights:

    .. math::

        w_i = 1 / (d(x, x_i)^{power} + r)

    By supplying a positive regularization parameter one can avoid
    singularities at the locations of the data points as well as control
    the "smoothness" of the interpolation (e.g., make the weights of the
    neighbors less varied). The "smoothness" of interpolation can also
    be controlled by the power parameter (``power``).

    Examples
    --------
    This class can be instantiated using the following syntax::

        >>> from photutils.utils import ShepardIDWInterpolator as idw

    Example of interpolating 1D data::

        >>> import numpy as np
        >>> rng = np.random.default_rng(0)
        >>> x = rng.random(100)  # 100 random values
        >>> y = np.sin(x)
        >>> f = idw(x, y)
        >>> float(f(0.4))  # doctest: +FLOAT_CMP
        0.38937843420912366
        >>> float(np.sin(0.4))  # doctest: +FLOAT_CMP
        0.3894183423086505

        >>> xi = rng.random(4)  # 4 random values
        >>> xi  # doctest: +FLOAT_CMP
        array([0.47998792, 0.23237292, 0.80188058, 0.92353016])
        >>> f(xi)  # doctest: +FLOAT_CMP
        array([0.46577097, 0.22837422, 0.71856662, 0.80125391])
        >>> np.sin(xi)  # doctest: +FLOAT_CMP
        array([0.46176846, 0.23028731, 0.71866503, 0.7977353 ])

    NOTE: In the last example, ``xi`` may be a ``Nx1`` array instead of
    a 1D vector.

    Example of interpolating 2D data::

        >>> rng = np.random.default_rng(0)
        >>> pos = rng.random((1000, 2))
        >>> val = np.sin(pos[:, 0] + pos[:, 1])
        >>> f = idw(pos, val)
        >>> float(f([0.5, 0.6]))  # doctest: +FLOAT_CMP
        0.8948257014687874
        >>> float(np.sin(0.5 + 0.6))  # doctest: +FLOAT_CMP
        0.8912073600614354
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, coordinates, values, weights=None, leafsize=10):
        coordinates = np.asarray(coordinates)
        if coordinates.ndim == 0:  # scalar coordinate
            coordinates = np.atleast_2d(coordinates)

        if coordinates.ndim == 1:
            coordinates = np.transpose(np.atleast_2d(coordinates))

        if coordinates.ndim > 2:
            coordinates = np.reshape(coordinates, (-1, coordinates.shape[-1]))

        values = np.asanyarray(values).ravel()

        ncoords = coordinates.shape[0]
        if ncoords < 1:
            msg = 'coordinates must have at least one data point'
            raise ValueError(msg)

        if values.shape[0] != ncoords:
            msg = 'The number of values must match the number of coordinates.'
            raise ValueError(msg)

        if weights is not None:
            weights = np.asanyarray(weights).ravel()
            if weights.shape[0] != ncoords:
                msg = ('The number of weights must match the number of '
                       'coordinates.')
                raise ValueError(msg)
            if np.any(weights < 0.0):
                msg = 'All weight values must be non-negative numbers.'
                raise ValueError(msg)

        self.coordinates = coordinates
        self.ncoords = ncoords
        self.coords_ndim = coordinates.shape[1]
        self.values = values
        self.weights = weights
        self.kdtree = cKDTree(coordinates, leafsize=leafsize)

    @deprecated_renamed_argument('reg', 'regularization', '3.0',
                                 until='4.0')
    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __call__(self, positions, n_neighbors=8, eps=0.0, power=1.0,
                 regularization=0.0, conf_dist=1.0e-12, dtype=float):
        """
        Evaluate the interpolator at the given positions.

        Parameters
        ----------
        positions : float, 1D array_like, or NxM array_like
            Coordinates of the position(s) at which the interpolator
            should be evaluated. In general, it is expected that these
            coordinates are in a form of an NxM-like array where N is
            the number of points and M is dimension of the coordinate
            space. When M=1 (1D space), then the ``positions`` parameter
            may be input as a 1D-like array or, if only one data
            point is available, ``positions`` can be a scalar number
            representing the 1D coordinate of the data point.

            .. note::
                If the dimensionality of the ``positions`` argument is
                larger than 2, e.g., if it is of the form N1 x N2 x N3 x
                ... x Nn x M, then it will be flattened to form an array
                of size NxM where N = N1 * N2 * ... * Nn.

            .. warning::
                The dimensionality of ``positions`` must match the
                dimensionality of the ``coordinates`` used during the
                initialization of the interpolator.

        n_neighbors : int, optional
            The maximum number of nearest neighbors to use during the
            interpolation.

        eps : float, optional
            Set to use approximate nearest neighbors; the kth neighbor
            is guaranteed to be no further than (1 + ``eps``) times the
            distance to the real *k*-th nearest neighbor. See
            `scipy.spatial.cKDTree.query` for further information.

        power : float, optional
            The power of the inverse distance used for the interpolation
            weights. See the Notes section for more details.

        regularization : float, optional
            The regularization parameter. It may be used to control the
            smoothness of the interpolator. See the Notes section for
            more details.

        conf_dist : float, optional
            The confusion distance below which the interpolator should
            use the value of the closest data point instead of
            attempting to interpolate. This is used to avoid
            singularities at the known data points, especially if
            ``regularization`` is 0.0.

        dtype : data-type, optional
            The data type of the output interpolated values. If `None`
            then the type will be inferred from the type of the
            ``values`` parameter used during the initialization of the
            interpolator.

        Returns
        -------
        result : float or `~numpy.ndarray`
            The interpolated value(s). A scalar is returned when a
            single position is provided; otherwise a 1D array is
            returned.
        """
        n_neighbors = int(n_neighbors)
        if n_neighbors < 1:
            msg = 'n_neighbors must be a positive integer'
            raise ValueError(msg)

        if conf_dist is not None and conf_dist <= 0.0:
            conf_dist = None

        positions = np.asanyarray(positions)
        if positions.ndim == 0:
            # Assume we have a single 1D coordinate
            if self.coords_ndim != 1:
                msg = ('The dimensionality of the input position does '
                       'not match the dimensionality of the coordinates '
                       'used to initialize the interpolator.')
                raise ValueError(msg)
        elif positions.ndim == 1:
            # Assume we have a single point
            if self.coords_ndim not in (1, positions.shape[-1]):
                msg = ('The input position was provided as a 1D array, '
                       'but its length does not match the dimensionality '
                       'of the coordinates used to initialize the '
                       'interpolator.')
                raise ValueError(msg)
        elif positions.ndim != 2:
            msg = ('The input positions must be an array_like object '
                   'of dimensionality no larger than 2.')
            raise ValueError(msg)

        positions = np.reshape(positions, (-1, self.coords_ndim))
        n_positions = positions.shape[0]

        distances, idx = self.kdtree.query(positions, k=n_neighbors, eps=eps)

        if n_neighbors == 1:
            result = self.values[idx]
            return result.item() if n_positions == 1 else result

        if dtype is None:
            dtype = self.values.dtype

        # distances and idx have shape (n_positions, n_neighbors). Mask
        # for valid (finite) distances; invalid entries arise when
        # n_neighbors exceeds the number of known data points.
        valid = np.isfinite(distances)

        # Replace non-finite distances and out-of-bound indices with
        # safe values so that vectorized indexing and arithmetic do not
        # raise errors; the ``valid`` mask zeroes them out later.
        safe_distances = np.where(valid, distances, 1.0)
        safe_idx = np.where(valid, idx, 0)

        # Inverse distance weights: w_i = 1 / (d_i^power + reg) The
        # errstate context suppresses divide-by-zero warnings that occur
        # when a query point coincides with a data point (distance = 0
        # and reg = 0); these are handled by the conf_dist override.
        with np.errstate(invalid='ignore', divide='ignore'):
            weights = np.where(valid,
                               1.0 / (safe_distances ** power
                                      + regularization),
                               0.0)

            # Apply external (user-supplied) weights
            if self.weights is not None:
                weights *= np.where(valid, self.weights[safe_idx], 0.0)

            # Gather neighbor values and compute the weighted average
            neighbor_values = self.values[safe_idx]
            weights_tot = np.sum(weights, axis=1)
            weighted_sum = np.sum(weights * neighbor_values, axis=1)

            # Where total weight is positive, compute interpolation;
            # otherwise return NaN (covers both the "no valid
            # neighbours" and "all-zero external weights" cases).
            interp_values = np.where(weights_tot > 0.0,
                                     weighted_sum / weights_tot,
                                     np.nan).astype(dtype)

        # Confusion-distance override: if the nearest neighbour is
        # closer than ``conf_dist``, return its value directly instead
        # of interpolating (avoids singularities when reg == 0).
        if conf_dist is not None:
            min_dist = distances[:, 0]
            confused = np.isfinite(min_dist) & (min_dist <= conf_dist)
            if np.any(confused):
                interp_values[confused] = self.values[
                    idx[confused, 0]
                ].astype(dtype)

        if n_positions == 1:
            return interp_values[0]

        return interp_values
